from samplerate import AudioSampleRate, audio_sample_rate
from soundfile import SoundFile
from io import BytesIO
from zounds.core import IdentityDimension, ArrayWithUnits
from timeseries import TimeDimension, Seconds
from duration import Picoseconds
from samplerate import SampleRate
import numpy as np


class AudioSamples(ArrayWithUnits):
    """
    `AudioSamples` represents constant-rate samples of a continuous audio signal
    at common sampling rates.

    It is a special case of an :class:`~zounds.core.ArrayWithUnits` whose first
    dimension is a :class:`~zounds.timeseries.TimeDimension` that has a common
    audio sampling rate (e.g. :class:`~zounds.timeseries.SR44100`).

    Args:
        array (np.ndarray): The raw sample data
        samplerate (SampleRate): The rate at which data was sampled

    Raises:
        ValueError: When array has a second dimension with size greater than 2
        TypeError: When samplerate is not a
            :class:`~zounds.timeseries.AudioSampleRate`
            (e.g. :class:`~zounds.timeseries.SR22050`)

    Examples::
        >>> from zounds import AudioSamples, SR44100, TimeSlice, Seconds
        >>> import numpy as np
        >>> raw = np.random.normal(0, 1, 44100*10)
        >>> samples = AudioSamples(raw, SR44100())
        >>> samples.samples_per_second
        44100
        >>> samples.channels
        1
        >>> sliced = samples[TimeSlice(Seconds(2))]
        >>> sliced.shape
        (88200,)
    """

    def __new__(cls, array, samplerate):
        if array.ndim == 1:
            dimensions = [TimeDimension(*samplerate)]
        elif array.ndim == 2:
            dimensions = [TimeDimension(*samplerate), IdentityDimension()]
        else:
            raise ValueError(
                'array must be one (mono) or two (multi-channel) dimensions')

        if not isinstance(samplerate, AudioSampleRate):
            raise TypeError('samplerate should be an AudioSampleRate instance')

        return ArrayWithUnits.__new__(cls, array, dimensions)

    def __add__(self, other):
        try:
            if self.samplerate != other.samplerate:
                raise ValueError(
                    'Samplerates must match, but they were '
                    '{self.samplerate} and {other.samplerate}'
                        .format(**locals()))
        except AttributeError:
            pass
        return super(AudioSamples, self).__add__(other)

    def kwargs(self):
        return {'samplerate': self.samplerate}

    def sum(self, axis=None, dtype=None, **kwargs):
        result = super(AudioSamples, self).sum(axis, dtype, **kwargs)
        if self.ndim == 2 and axis == 1:
            return AudioSamples(result, self.samplerate)
        else:
            return result

    @classmethod
    def silence(cls, samplerate, duration, dtype=np.float32):
        discrete_samples = int(duration / samplerate.frequency)
        silence = np.zeros(discrete_samples, dtype=dtype)
        return cls(silence, samplerate)

    def pad_with_silence(self, silence_duration):
        silence = self.__class__.silence(
            self.samplerate, silence_duration, self.dtype)
        return AudioSamples(np.concatenate([self, silence]), self.samplerate)

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def duration_in_seconds(self):
        return self.duration / Picoseconds(int(1e12))

    @property
    def samplerate(self):
        return SampleRate(self.frequency, self.duration)

    @property
    def overlap(self):
        return self.samplerate.overlap

    @property
    def span(self):
        return self.dimensions[0].span

    @property
    def end(self):
        return self.dimensions[0].end

    @property
    def frequency(self):
        return self.dimensions[0].frequency

    @property
    def duration(self):
        return self.dimensions[0].duration

    @classmethod
    def from_example(cls, arr, example):
        return cls(arr, example.samplerate)

    @property
    def channels(self):
        if len(self.shape) == 1:
            return 1
        return self.shape[1]

    @property
    def samplerate(self):
        return audio_sample_rate(self.samples_per_second)

    @property
    def mono(self):
        """
        Return this instance summed to mono.  If the instance is already mono,
        this is a no-op.
        """
        if self.channels == 1:
            return self
        x = self.sum(axis=1) * 0.5
        y = x * 0.5
        return AudioSamples(y, self.samplerate)

    @property
    def stereo(self):
        if self.channels == 2:
            return self
        return AudioSamples(np.vstack([self, self]).T, self.samplerate)

    def encode(self, flo=None, fmt='WAV', subtype='PCM_16'):
        """
        Return audio samples encoded as bytes given a particular audio format

        Args:
            flo (file-like): A file-like object to write the bytes to.  If flo
                is not supplied, a new :class:`io.BytesIO` instance will be
                 created and returned
            fmt (str): A libsndfile-friendly identifier for an audio encoding
                (detailed here: http://www.mega-nerd.com/libsndfile/api.html)
            subtype (str): A libsndfile-friendly identifier for an audio
                encoding subtype (detailed here:
                http://www.mega-nerd.com/libsndfile/api.html)

        Examples:
            >>> from zounds import SR11025, AudioSamples
            >>> import numpy as np
            >>> silence = np.zeros(11025*10)
            >>> samples = AudioSamples(silence, SR11025())
            >>> bio = samples.encode()
            >>> bio.read(10)
            'RIFFx]\\x03\\x00WA'
        """
        flo = flo or BytesIO()
        with SoundFile(
                flo,
                mode='w',
                channels=self.channels,
                format=fmt,
                subtype=subtype,
                samplerate=self.samples_per_second) as f:
            f.write(self)
        flo.seek(0)
        return flo



"""
The `synthesize` module includes classes that can produce audio.  Some, like
:class:`SineSynthesize` can produce simple signals from scratch that are often
useful for test-cases, while others are able to invert common frequency-domain
transforms, like the :class:`MDCTSynthesizer`
"""

from synthesize import \
    FFTSynthesizer, DCTSynthesizer, TickSynthesizer, NoiseSynthesizer, \
    SineSynthesizer, DCTIVSynthesizer, MDCTSynthesizer, \
    FrequencyAdaptiveFFTSynthesizer, FrequencyAdaptiveDCTSynthesizer, \
    SilenceSynthesizer, WindowedAudioSynthesizer, \
    FrequencyDecompositionSynthesizer



from resample import Resample
from zounds.timeseries import SR44100, SR11025, Seconds
from zounds.synthesize import SilenceSynthesizer
from multiprocessing.pool import ThreadPool
import unittest2


class ResampleTests(unittest2.TestCase):

    def test_can_do_multithreaded_resampling(self):
        synth = SilenceSynthesizer(SR44100())
        audio = [synth.synthesize(Seconds(5)) for _ in xrange(10)]
        pool = ThreadPool(4)

        def x(samples):
            rs = Resample(int(SR44100()), int(SR11025()))
            return rs(samples, end_of_input=True)

        resampled = pool.map(x, audio)
        self.assertEqual(10, len(resampled))


from __future__ import division
from frequencyscale import LinearScale, FrequencyBand, ExplicitScale
from tfrepresentation import FrequencyDimension
from frequencyadaptive import FrequencyAdaptive
from zounds.timeseries import audio_sample_rate, TimeSlice
from zounds.core import ArrayWithUnits
from sliding_window import IdentityWindowingFunc
import numpy as np
from scipy.signal import resample


def fft(x, axis=-1):
    transformed = np.fft.rfft(x, axis=axis, norm='ortho')

    sr = audio_sample_rate(
        int(x.shape[1] / x.dimensions[0].duration_in_seconds))
    scale = LinearScale.from_sample_rate(sr, transformed.shape[-1])

    return ArrayWithUnits(
        transformed, [x.dimensions[0], FrequencyDimension(scale)])


def stft(x, window_sample_rate, window=None):
    duration = TimeSlice(window_sample_rate.duration)
    frequency = TimeSlice(window_sample_rate.frequency)
    _, arr = x.sliding_window_with_leftovers(
        duration, frequency, dopad=True)
    window = window or IdentityWindowingFunc()
    windowed = arr * window
    return fft(windowed)


def apply_scale(short_time_fft, scale, reducer=np.sum, window=None):
    magnitudes = np.abs(short_time_fft.real)
    output = np.zeros(
        short_time_fft.shape[:-1] + (len(scale),), dtype=magnitudes.dtype)
    output = ArrayWithUnits(
        output, short_time_fft.dimensions[:-1] + (FrequencyDimension(scale),))
    window = window or IdentityWindowingFunc()
    for i, freq_band in enumerate(scale):
        reduced_band = reducer(magnitudes[..., freq_band] * window, axis=-1)
        output[..., i] = reduced_band
    return output


def frequency_decomposition(x, sizes):
    sizes = sorted(sizes)

    original_size = x.shape[-1]
    time_dimension = x.dimensions[-1]
    samplerate = audio_sample_rate(time_dimension.samples_per_second)
    data = x.copy()

    bands = []
    frequency_bands = []
    start_hz = 0

    for size in sizes:
        if size != original_size:
            s = resample(data, size, axis=-1)
        else:
            s = data

        bands.append(s)
        data -= resample(s, original_size, axis=-1)

        stop_hz = samplerate.nyquist * (size / original_size)
        frequency_bands.append(FrequencyBand(start_hz, stop_hz))
        start_hz = stop_hz

    scale = ExplicitScale(frequency_bands)
    return FrequencyAdaptive(bands, scale=scale, time_dimension=x.dimensions[0])



from zounds.soundfile import AudioMetaData
import requests
import re
import urlparse


class PhatDrumLoops(object):
    """
    Produces an iterable of :class:`zounds.soundfile.AudioMetaData` instances
    for every drum break from http://phatdrumloops.com/beats.php

    Args:
        attrs (dict): Extra properties to add to the :class:`AudioMetaData`

    Examples
        >>> from zounds import PhatDrumLoops
        >>> pdl = PhatDrumLoops()
        >>> iter(pdl).next()
        {'description': None, 'tags': None, 'uri': <Request [GET]>, 'channels': None, 'licensing': None, 'samplerate': None}


    See Also:
        :class:`InternetArchive`
        :class:`FreeSoundSearch`
        :class:`zounds.soundfile.AudioMetaData`
    """
    def __init__(self, **attrs):
        super(PhatDrumLoops, self).__init__()
        self.attrs = attrs
        self.attrs.update(web_url='http://www.phatdrumloops.com/beats.php')

    def __iter__(self):
        resp = requests.get('http://phatdrumloops.com/beats.php')
        pattern = re.compile('href="(?P<uri>/audio/wav/[^\.]+\.wav)"')
        for m in pattern.finditer(resp.content):
            url = urlparse.urljoin('http://phatdrumloops.com',
                                   m.groupdict()['uri'])
            request = requests.Request(
                method='GET',
                url=url,
                headers={'Range': 'bytes=0-'})
            yield AudioMetaData(uri=request, **self.attrs)



from __future__ import division
from featureflow import IdentityEncoder, Node, Decoder, Feature
from audiostream import MemoryBuffer
from zounds.timeseries import audio_sample_rate, TimeSlice, AudioSamples
from soundfile import *
from byte_depth import chunk_size_samples
from zounds.timeseries import Picoseconds, Seconds


class OggVorbisWrapper(object):
    def __init__(self, flo):
        self._flo = flo
        self._sf = SoundFile(self._flo)
        self._freq = Picoseconds(int(1e12)) / self._sf.samplerate

    @property
    def samplerate(self):
        return self._sf.samplerate

    @property
    def channels(self):
        return self._sf.channels

    @property
    def duration_seconds(self):
        return len(self._sf) / self.samplerate

    def _n_samples(self, duration):
        if duration is None:
            return -1
        return int(duration / self._freq)

    def __getitem__(self, timeslice):
        sr = audio_sample_rate(self.samplerate)

        if timeslice == slice(None):
            self._sf.seek(0)
            return AudioSamples(self._sf.read(len(self._sf)), sr)

        start_sample = int(timeslice.start / self._freq)
        self._sf.seek(start_sample)
        n_samples = self._n_samples(timeslice.duration)
        return AudioSamples(self._sf.read(n_samples), sr)

    def iter_chunks(self):
        chunksize = Seconds(1)
        ts = TimeSlice(chunksize)
        sl = self[ts]
        yield sl
        while len(sl) >= self._n_samples(chunksize):
            ts += chunksize
            sl = self[ts]
            yield sl


class OggVorbisEncoder(IdentityEncoder):
    content_type = 'audio/ogg'


class OggVorbisDecoder(Decoder):
    def __init__(self):
        super(OggVorbisDecoder, self).__init__()

    def __call__(self, flo):
        return OggVorbisWrapper(flo)

    def __iter__(self, flo):
        yield self(flo)


class OggVorbisFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            **extractor_args):
        super(OggVorbisFeature, self).__init__( \
                extractor,
                needs=needs,
                store=store,
                encoder=OggVorbisEncoder,
                decoder=OggVorbisDecoder(),
                key=key,
                **extractor_args)


class OggVorbis(Node):
    """
    `OggVorbis` expects to process a stream of raw bytes (e.g. one produced by
    :class:`featureflow.ByteStream`) and produces a new byte stream where the
    original audio samples are `ogg-vorbis <https://xiph.org/vorbis/>`_ encoded

    Args:
        needs (Feature): a feature that produces a byte stream
            (e.g. :class:`featureflow.Bytestream`)

    Here's how you'd typically see :class:`OggVorbis` used in a processing
    graph.

    .. code:: python

        import featureflow as ff
        import zounds


        chunksize = zounds.ChunkSizeBytes(
            samplerate=zounds.SR44100(),
            duration=zounds.Seconds(30),
            bit_depth=16,
            channels=2)

        @zounds.simple_in_memory_settings
        class Document(ff.BaseModel):
            meta = ff.JSONFeature(
                zounds.MetaData,
                store=True,
                encoder=zounds.AudioMetaDataEncoder)

            raw = ff.ByteStreamFeature(
                ff.ByteStream,
                chunksize=chunksize,
                needs=meta,
                store=False)

            ogg = zounds.OggVorbisFeature(
                zounds.OggVorbis,
                needs=raw,
                store=True)


        synth = zounds.NoiseSynthesizer(zounds.SR11025())
        samples = synth.synthesize(zounds.Seconds(10))
        raw_bytes = samples.encode()
        _id = Document.process(meta=raw_bytes)
        doc = Document(_id)
        # fetch and decode a section of audio
        ts = zounds.TimeSlice(zounds.Seconds(2))
        print doc.ogg[ts].shape  # 22050
    """
    def __init__(self, needs=None):
        super(OggVorbis, self).__init__(needs=needs)
        self._in_buf = None
        self._in_sf = None
        self._out_buf = None
        self._out_sf = None
        self._already_ogg = None
        self._chunk_size_samples = None

    def _enqueue(self, data, pusher):
        if self._in_buf is None:
            self._in_buf = MemoryBuffer(data.total_length)
            self._in_buf.write(data)
            self._in_sf = SoundFile(self._in_buf)
            self._already_ogg = 'OGG' in self._in_sf.format

        if not self._chunk_size_samples:
            self._chunk_size_samples = chunk_size_samples(self._in_sf, data)

        if self._already_ogg:
            super(OggVorbis, self)._enqueue(data, pusher)
            return

        if self._out_buf is None:
            self._out_buf = MemoryBuffer(data.total_length)
            self._out_sf = SoundFile(
                    self._out_buf,
                    format='OGG',
                    subtype='VORBIS',
                    mode='w',
                    samplerate=self._in_sf.samplerate,
                    channels=self._in_sf.channels)
        else:
            self._in_buf.write(data)

    def _dequeue(self):

        if self._already_ogg:
            return super(OggVorbis, self)._dequeue()

        samples = self._in_sf.read(self._chunk_size_samples)
        factor = 20
        while samples.size:
            # KLUDGE: Trying to write too-large chunks to an ogg file seems to
            # cause a segfault in libsndfile
            for i in xrange(0, len(samples), self._in_sf.samplerate * factor):
                self._out_sf.write(
                        samples[i: i + self._in_sf.samplerate * factor])
            samples = self._in_sf.read(self._chunk_size_samples)
        return self._out_buf

    def _process_other(self, data):
        if self._finalized:
            self._out_sf.close()
        o = data.read(count=-1)
        return o

    def _process_ogg(self, data):
        return data

    def _process(self, data):
        if self._already_ogg:
            yield self._process_ogg(data)
        else:
            yield self._process_other(data)



import unittest2
from zounds.util import simple_in_memory_settings
from preprocess import MeanStdNormalization, PreprocessingPipeline
import featureflow as ff
import numpy as np


class MeanStdTests(unittest2.TestCase):
    def _forward_backward(self, shape):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            meanstd = ff.PickleFeature(
                    MeanStdNormalization,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(meanstd,),
                    store=True)

        training = np.random.random_sample((100,) + shape)
        _id = Model.process(meanstd=training)
        model = Model(_id)

        data_shape = (10,) + shape
        data = np.random.random_sample(data_shape)
        result = model.pipeline.transform(data)
        self.assertEqual(data_shape, result.data.shape)
        inverted = result.inverse_transform()
        self.assertEqual(inverted.shape, data.shape)
        np.testing.assert_allclose(inverted, data)

    def test_can_process_1d(self):
        self._forward_backward((9,))

    def test_can_process_2d(self):
        self._forward_backward((3, 4))

    def test_can_process_3d(self):
        self._forward_backward((5, 4, 7))



from featureflow import Feature
from arraywithunits import ArrayWithUnitsDecoder, ArrayWithUnitsEncoder
from zounds.spectral import FrequencyAdaptive


class FrequencyAdaptiveDecoder(ArrayWithUnitsDecoder):
    def __init__(self):
        super(ArrayWithUnitsDecoder, self).__init__()

    def __call__(self, flo):
        raw = super(FrequencyAdaptiveDecoder, self).__call__(flo)
        return FrequencyAdaptive.from_array_with_units(raw)


class FrequencyAdaptiveFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=ArrayWithUnitsEncoder,
            decoder=FrequencyAdaptiveDecoder(),
            **extractor_args):
        super(FrequencyAdaptiveFeature, self).__init__(
            extractor,
            needs=needs,
            store=store,
            encoder=encoder,
            decoder=decoder,
            key=key,
            **extractor_args)



from __future__ import division

import numpy as np
from scipy.signal import resample

from tfrepresentation import ExplicitFrequencyDimension, FrequencyDimension
from zounds.core import ArrayWithUnits
from zounds.timeseries import ConstantRateTimeSeries
from zounds.timeseries import Picoseconds, TimeDimension


class FrequencyAdaptive(ArrayWithUnits):
    """
    TODO: This needs some love. Mutually exclusive constructor arguments are no
    bueno

    Args:
        arrs: TODO
        time_dimension (TimeDimension): the time dimension of the first axis of
            this array
        scale (FrequencyScale): The frequency scale corresponding to the first
            axis of this array, mutually exclusive with the
            :code:`explicit_freq_dimension` argument
        explicit_freq_dimension (ExplicitFrequencyDimension): TODO

    See Also:
        :class:`~zounds.spectral.FrequencyAdaptiveTransform`
    """
    def __new__(
            cls,
            arrs,
            time_dimension=None,
            scale=None,
            explicit_freq_dimension=None):

        if not time_dimension:
            raise ValueError('time_dimension is required')

        if explicit_freq_dimension:
            if scale:
                raise ValueError(
                    'scale must be None when explicit_freq_dimension is supplied')
            if not isinstance(arrs, np.ndarray):
                raise ValueError(
                    'arrs must be a contiguous array when explicit_freq_dimension_is_supplied')
            return ArrayWithUnits.__new__(
                cls, arrs, [time_dimension, explicit_freq_dimension])

        stops = list(np.cumsum([arr.shape[1] for arr in arrs]))
        slices = [slice(start, stop)
                  for (start, stop) in zip([0] + stops, stops)]
        dimensions = [time_dimension, ExplicitFrequencyDimension(scale, slices)]

        array = np.concatenate(arrs, axis=1)
        return ArrayWithUnits.__new__(cls, array, dimensions)

    def kwargs(self):
        return dict(
            time_dimension=self.time_dimension,
            explicit_freq_dimension=self.frequency_dimension)

    @property
    def scale(self):
        return self.frequency_dimension.scale

    @property
    def time_dimension(self):
        return self.dimensions[0]

    @property
    def frequency_dimension(self):
        return self.dimensions[1]

    @property
    def n_bands(self):
        return len(self.scale)

    def rasterize(self, n_coeffs):
        return self.square(n_coeffs)

    def _resample(self, band, n_coeffs):
        rs = resample(band, n_coeffs, axis=1)

        # resample doesn't necessarily maintain the correct scale/magnitude, as
        # it isn't using the norm="ortho" argument when calling fft, so ensure
        # that the original scale/magnitude is maintained after calling
        # resample
        m = band.max(axis=-1, keepdims=True)
        ratio = np.divide(
            rs.max(axis=-1, keepdims=True), m, where=m != 0)
        normalized = np.divide(rs, ratio, where=ratio != 0)
        return np.asarray(normalized.flatten())

    def square(self, n_coeffs, do_overlap_add=False):
        """
        Compute a "square" view of the frequency adaptive transform, by
        resampling each frequency band such that they all contain the same
        number of samples, and performing an overlap-add procedure in the
        case where the sample frequency and duration differ
        :param n_coeffs: The common size to which each frequency band should
        be resampled
        """
        resampled_bands = [
            self._resample(band, n_coeffs)
            for band in self.iter_bands()]

        stacked = np.vstack(resampled_bands).T

        fdim = FrequencyDimension(self.scale)

        # TODO: This feels like it could be wrapped up nicely elsewhere
        chunk_frequency = Picoseconds(int(np.round(
            self.time_dimension.duration / Picoseconds(1) / n_coeffs)))

        td = TimeDimension(frequency=chunk_frequency)

        arr = ConstantRateTimeSeries(ArrayWithUnits(
            stacked.reshape(-1, n_coeffs, self.n_bands),
            dimensions=[self.time_dimension, td, fdim]))

        if not do_overlap_add:
            return arr

        # Begin the overlap add procedure
        overlap_ratio = self.time_dimension.overlap_ratio

        if overlap_ratio == 0:
            # no overlap add is necessary
            return ArrayWithUnits(stacked, [td, fdim])

        step_size_samples = int(n_coeffs * overlap_ratio)

        first_dim = int(np.round(
            (stacked.shape[0] * overlap_ratio) + (n_coeffs * overlap_ratio)))

        output = ArrayWithUnits(
            np.zeros((first_dim, self.n_bands)),
            dimensions=[td, fdim])

        for i, chunk in enumerate(arr):
            start = step_size_samples * i
            stop = start + n_coeffs
            output[start: stop] += chunk.reshape((-1, self.n_bands))

        return output

    def iter_bands(self):
        return (self[:, band] for band in self.scale)

    def like_dims(self, arr):
        return self.__class__(
            arr,
            time_dimension=self.time_dimension,
            explicit_freq_dimension=self.frequency_dimension)

    @classmethod
    def from_array_with_units(cls, arr):
        fdim = arr.dimensions[1]
        arrs = [arr[:, band] for band in fdim.scale]
        fa = FrequencyAdaptive(arrs, arr.dimensions[0], fdim.scale)
        return fa



from multiprocessing.pool import ThreadPool, Pool, cpu_count
from itertools import repeat, izip, imap


def ingest_one(arg):
    metadata, cls, skip_if_exists = arg
    request = metadata.request
    url = request.url
    if skip_if_exists and cls.exists(request.url):
        print 'already processed {request.url}'.format(**locals())
        return

    try:
        print 'processing {request.url}'.format(**locals())
        cls.process(meta=metadata, _id=url)
    except Exception as e:
        print e


def ingest(
        dataset,
        cls,
        skip_if_exists=True,
        multi_process=False,
        multi_threaded=False,
        cores=None):

    pool = None

    if multi_process:
        pool = Pool(cores or cpu_count())
        map_func = pool.imap_unordered
    elif multi_threaded:
        pool = ThreadPool(cores or cpu_count())
        map_func = pool.imap_unordered
    else:
        map_func = map

    cls_args = repeat(cls)
    skip_args = repeat(skip_if_exists)

    map_func(ingest_one, izip(dataset, cls_args, skip_args))

    if pool is not None:
        # if we're ingesting using multiple processes or threads, the processing
        # should be parallel, but this method should be synchronous from the
        # caller's perspective
        pool.close()
        pool.join()




"""
The `learn` module includes classes that make it possible to define processing
graphs whose leaves are trained machine learning models.

While much of :module:`zounds.soundfile`, :module:`zounds.spectral`, and
:module:`zounds.timeseries` focus on processing nodes that can be composed into
a processing graph to extract features from a single piece of audio, the `learn`
module focuses on defining graphs that extract features or trained models from
an entire corpus of audio.
"""

from learn import KMeans, Learned

from preprocess import \
    MeanStdNormalization, UnitNorm, Log, PreprocessingPipeline, Multiply, \
    Slicer, InstanceScaling, Reshape, Weighted, MuLawCompressed, SimHash, \
    AbsoluteValue, Binarize, Sharpen

from sklearn_preprocessor import SklearnModel, WithComponents

from pytorch_model import PyTorchAutoEncoder, PyTorchGan, PyTorchNetwork

from gan import GanTrainer
from wgan import WassersteinGanTrainer
from supervised import SupervisedTrainer
from embedding import TripletEmbeddingTrainer

from random_samples import ReservoirSampler, ShuffledSamples

from util import simple_settings

from graph import learning_pipeline

from functional import hyperplanes, simhash, example_wise_unit_norm

try:
    from util import \
        Conv1d, ConvTranspose1d, Conv2d, ConvTranspose2d, to_var, from_var, \
        try_network, apply_network, feature_map_size, \
        FrequencyDecompositionAnalyzer, FrequencyDecompositionGenerator
except ImportError:
    pass





import base64
import re
import numpy as np

TIMEDELTA_DTYPE_RE = re.compile(r'\[(?P<dtype>[^\]]+)\]')


def encode_timedelta(td):
    dtype = TIMEDELTA_DTYPE_RE.search(str(td.dtype)).groupdict()['dtype']
    return base64.b64encode(td.astype(np.uint64).tostring()), dtype


def decode_timedelta(t):
    if isinstance(t, np.timedelta64):
        return t

    v = np.fromstring(base64.b64decode(t[0]), dtype=np.uint64)[0]
    s = t[1]
    return np.timedelta64(long(v), s)



from featureflow import BaseModel
from sliding_window import \
    SlidingWindow, IdentityWindowingFunc, OggVorbisWindowingFunc
from zounds.timeseries import \
    AudioSamples, SR22050, SR44100, SR11025, SR48000, SR96000, SampleRate, \
    Picoseconds, Seconds, Milliseconds, TimeDimension, HalfLapped, TimeSlice
from zounds.util import simple_in_memory_settings
from zounds.basic import resampled
from zounds.synthesize import NoiseSynthesizer
from zounds.core import ArrayWithUnits
from zounds.persistence import ArrayWithUnitsFeature
from zounds.spectral import FrequencyDimension, LinearScale, FrequencyBand
import numpy as np
import unittest2


class WindowingFunctionTests(unittest2.TestCase):
    def test_multiply_lhs(self):
        samples = np.random.random_sample(44)
        wf = IdentityWindowingFunc()
        np.testing.assert_allclose(wf * samples, samples)

    def test_multiply_rhs(self):
        samples = np.random.random_sample(33)
        wf = IdentityWindowingFunc()
        np.testing.assert_allclose(samples * wf, samples)


class OggVorbisWindowingFunctionTests(unittest2.TestCase):
    def test_multilpy_many_frames(self):
        samples = np.random.random_sample((10, 3))
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)

    def test_multiply_one_frame(self):
        samples = np.random.random_sample((10, 1))
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)

    def test_multiply_1d(self):
        samples = np.random.random_sample(10)
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)


class SlidingWindowTests(unittest2.TestCase):
    def _check(self, samplerate, expected_window_size, expected_step_size):
        # sw = SlidingWindow(wscheme=samplerate.half_lapped())
        samples = AudioSamples(
            np.zeros(5 * samplerate.samples_per_second), samplerate)
        # sw._enqueue(samples, None)
        wscheme = samplerate.half_lapped()
        ws, ss = samples._sliding_window_integer_slices(
            TimeSlice(wscheme.duration), TimeSlice(wscheme.frequency))
        self.assertEqual(expected_window_size, ws[0])
        self.assertEqual(expected_step_size, ss[0])

    def test_correct_window_and_step_size_at_96000(self):
        self._check(SR96000(), 4096, 2048)

    def test_correct_window_and_step_size_at_48000(self):
        self._check(SR48000(), 2048, 1024)

    def test_correct_window_and_step_size_at_22050(self):
        self._check(SR22050(), 1024, 512)

    def test_correct_window_and_step_size_at_44100(self):
        self._check(SR44100(), 2048, 1024)

    def test_correct_window_and_step_size_at_11025(self):
        self._check(SR11025(), 512, 256)

    def test_can_persist_and_retrieve_with_second_long_windowing_scheme(self):
        samplerate = SR22050()
        rs = resampled(resample_to=samplerate)

        window_size = Seconds(1)
        wscheme = SampleRate(window_size, window_size)

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=wscheme,
                needs=rs.resampled,
                store=True)

        synth = NoiseSynthesizer(samplerate)
        audio = synth.synthesize(Milliseconds(5500))

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        self.assertEqual(6, len(doc.windowed))

    def test_has_correct_duration(self):
        samplerate = SR22050()
        rs = resampled(resample_to=samplerate)

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=HalfLapped(),
                needs=rs.resampled,
                store=True)

        synth = NoiseSynthesizer(samplerate)
        audio = synth.synthesize(Milliseconds(5500))

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        orig_seconds = audio.dimensions[0].end / Picoseconds(int(1e12))
        new_seconds = doc.windowed.dimensions[0].end / Picoseconds(int(1e12))
        self.assertAlmostEqual(orig_seconds, new_seconds, delta=0.01)

    def test_can_apply_sliding_window_to_constant_rate_time_series(self):
        arr = ArrayWithUnits(np.zeros(100), [TimeDimension(Seconds(1))])
        sw = SampleRate(Seconds(2), Seconds(2))

        @simple_in_memory_settings
        class Document(BaseModel):
            windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=sw,
                store=True)

        _id = Document.process(windowed=arr)
        result = Document(_id).windowed

        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((50, 2), result.shape)
        self.assertEqual(2, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], TimeDimension)
        self.assertEqual(Seconds(2), result.dimensions[0].frequency)
        self.assertIsInstance(result.dimensions[1], TimeDimension)
        self.assertEqual(Seconds(1), result.dimensions[1].frequency)

    def test_can_apply_sliding_window_to_time_frequency_representation(self):
        band = FrequencyBand(0, 22000)
        scale = LinearScale(band, 100)
        arr = ArrayWithUnits(
            np.zeros((200, 100)),
            [TimeDimension(Seconds(1)), FrequencyDimension(scale)])
        sw = SampleRate(Seconds(2), Seconds(2))

        @simple_in_memory_settings
        class Document(BaseModel):
            windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=sw,
                store=True)

        _id = Document.process(windowed=arr)
        result = Document(_id).windowed

        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((100, 2, 100), result.shape)
        self.assertEqual(3, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], TimeDimension)
        self.assertEqual(Seconds(2), result.dimensions[0].frequency)
        self.assertIsInstance(result.dimensions[1], TimeDimension)
        self.assertEqual(Seconds(1), result.dimensions[1].frequency)
        self.assertIsInstance(result.dimensions[2], FrequencyDimension)

    def test_sliding_window_maintains_dtype(self):
        band = FrequencyBand(0, 22000)
        scale = LinearScale(band, 100)
        arr = ArrayWithUnits(
            np.zeros((200, 100), dtype=np.uint8),
            [TimeDimension(Seconds(1)), FrequencyDimension(scale)])
        sw = SampleRate(Seconds(2), Seconds(2))

        @simple_in_memory_settings
        class Document(BaseModel):
            windowed = ArrayWithUnitsFeature(
                SlidingWindow,
                wscheme=sw,
                store=True)

        _id = Document.process(windowed=arr)
        result = Document(_id).windowed
        self.assertEqual(np.uint8, result.dtype)



import numpy as np
import frequencyscale
from zounds.core import Dimension
from zounds.spectral.frequencyscale import ExplicitScale


class FrequencyDimension(Dimension):
    """
    When applied to an axis of :class:`~zounds.core.ArrayWithUnits`, that axis
    can be viewed as representing the energy present in a series of frequency
    bands

    Args:
        scale (FrequencyScale): A scale whose frequency bands correspond to the
            items along the frequency axis

    Examples:
        >>> from zounds import LinearScale, FrequencyBand, ArrayWithUnits
        >>> from zounds import FrequencyDimension
        >>> import numpy as np
        >>> band = FrequencyBand(20, 20000)
        >>> scale = LinearScale(frequency_band=band, n_bands=100)
        >>> raw = np.hanning(100)
        >>> arr = ArrayWithUnits(raw, [FrequencyDimension(scale)])
        >>> sliced = arr[FrequencyBand(100, 1000)]
        >>> sliced.shape
        (5,)
        >>> sliced.dimensions
        (FrequencyDimension(scale=LinearScale(band=FrequencyBand(
        start_hz=20.0,
        stop_hz=1019.0,
        center=519.5,
        bandwidth=999.0), n_bands=5)),)
    """
    def __init__(self, scale):
        super(FrequencyDimension, self).__init__()
        self.scale = scale

    def copy(self):
        return FrequencyDimension(self.scale)

    def weights(self, weights, arr, i):
        return weights

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        return FrequencyDimension(self.scale[index])

    def integer_based_slice(self, index):
        if not isinstance(index, frequencyscale.FrequencyBand):
            return index

        return self.scale.get_slice(index)

    def validate(self, size):
        """
        Ensure that the size of the dimension matches the number of bands in the
        scale

        Raises:
             ValueError: when the dimension size and number of bands don't match
        """
        msg = 'scale and array size must match, ' \
              'but were scale: {self.scale.n_bands},  array size: {size}'

        if size != len(self.scale):
            raise ValueError(msg.format(**locals()))

    def __eq__(self, other):
        return self.scale == other.scale

    def __str__(self):
        return 'FrequencyDimension(scale={self.scale})'.format(**locals())

    def __repr__(self):
        return self.__str__()


class ExplicitFrequencyDimension(Dimension):
    """
    A frequency dimension where the mapping from frequency bands to integer
    indices is provided explicitly, rather than computed

    Args:
        scale (ExplicitScale): the explicit frequency scale that defines how
            slices are extracted from this dimension
        slices (iterable of slices): An iterable of :class:`python.slice`
            instances which correspond to each frequency band from scale

    Raises:
        ValueError: when the number of slices and number of bands in scale don't
            match
    """

    def __init__(self, scale, slices):
        super(ExplicitFrequencyDimension, self).__init__()
        if len(scale) != len(slices):
            raise ValueError('scale and slices must have same length')

        self.scale = scale
        self.slices = slices
        self._lookup = dict(zip(self.scale, self.slices))

    def copy(self):
        return ExplicitFrequencyDimension(self.scale, self.slices)

    def weights(self, weights, arr, i):
        w = np.zeros(self.slices[-1].stop - self.slices[0].start)
        for weight, sl in zip(weights, self.slices):
            w[sl] = weight
        return w

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        if isinstance(index, frequencyscale.FrequencyBand):
            try:
                slce = self._lookup[index]
                return ExplicitFrequencyDimension(
                    ExplicitScale([index]), [slce])
            except KeyError:
                slce = self.scale.get_slice(index)
        else:
            slce = index

        print slce
        return ExplicitFrequencyDimension(self.scale[slce], self.slices[slce])

    def integer_based_slice(self, index):
        if not isinstance(index, frequencyscale.FrequencyBand):
            return index

        try:
            return self._lookup[index]
        except KeyError:
            pass

        slce = self.scale.get_slice(index)
        slices = self.slices[slce]
        return slice(slices[0].start, slices[-1].stop)

    def validate(self, size):
        return size == self.slices[-1].stop

    def __eq__(self, other):
        return self.scale == other.scale and self.slices == other.slices

    def __str__(self):
        return \
            'ExplicitFrequencyDimension(scale={self.scale}, slices=self.slices)' \
                .format(**locals())

    def __repr__(self):
        return self.__str__()



from __future__ import division

from ctypes import *

import numpy as np

from zounds.timeseries import SR44100, AudioSamples
from zounds.core import ArrayWithUnits

try:
    libsamplerate = CDLL('libsamplerate.so')
except OSError, e:
    # KLUDGE: This is here to support building documentation on readthedocs
    pass

from featureflow import Node


class SRC_DATA(Structure):
    """
    A wrapper for the libsamplerate.SRC_DATA struct
    """
    _fields_ = [('data_in', POINTER(c_float)),
                ('data_out', POINTER(c_float)),
                ('input_frames', c_long),
                ('output_frames', c_long),
                ('input_frames_used', c_long),
                ('output_frames_gen', c_long),
                ('end_of_input', c_int),
                ('src_ratio', c_double), ]


class SRC_STATE(Structure):
    """
    A dummy structure to represent the state returned from libsamplerate
    src_new.
    """
    _fields_ = []


class Resample(object):
    """
    A wrapper around the libsamplerate src_process() method.  This class is
    intended for one-time use. New instances should be created for each sound\
    file processed.
    """

    def __init__(
            self,
            orig_sample_rate,
            new_sample_rate,
            nchannels=1,
            converter_type=1):

        """
        orig_sample_rate - The sample rate of the incoming samples, in hz
        new_sample_rate - The sample_rate of the outgoiing samples, in hz
        n_channels - Number of channels in the incoming and outgoing samples
        converter_type - See http://www.mega-nerd.com/SRC/api_misc.html#Converters
                         for a list of conversion types. "0" is the best-quality,
                         and slowest converter

        """
        super(Resample, self).__init__()
        self._ratio = new_sample_rate / orig_sample_rate
        # check if the conversion ratio is considered valid by libsamplerate
        if not libsamplerate.src_is_valid_ratio(c_double(self._ratio)):
            raise ValueError('%1.2f / %1.2f = %1.4f is not a valid ratio' % \
                             (new_sample_rate, orig_sample_rate, self._ratio))
        # create a pointer to the SRC_STATE struct, which maintains state
        # between calls to src_process()
        self.error = pointer(c_int(0))
        self.nchannels = nchannels
        self.converter_type = converter_type
        self.c_int_converter_type = c_int(converter_type)
        self.c_int_channels = c_int(self.nchannels)
        libsamplerate.src_new.restype = POINTER(SRC_STATE)
        self._state = libsamplerate.src_new(
            self.c_int_converter_type, self.c_int_channels, self.error)

    def _prepare_input(self, insamples):
        # ensure that the input is float data
        if np.float32 != insamples.dtype:
            return insamples.astype(np.float32)
        return insamples

    def _output_buffer(self, insamples):
        outsize = int(np.round(insamples.size * self._ratio))
        return np.zeros(outsize, dtype=np.float32)

    def _check_for_error(self, return_code):
        if return_code:
            raise Exception(
                c_char_p(
                    libsamplerate.src_strerror(c_int(return_code))).value)

    def __call__(self, insamples, end_of_input=False):

        normalized_insamples = self._prepare_input(insamples)
        outsamples = self._output_buffer(normalized_insamples)

        insamples_ptr = normalized_insamples.ctypes.data_as(POINTER(c_float))
        outsamples_ptr = outsamples.ctypes.data_as(POINTER(c_float))

        sd = SRC_DATA(
            # a pointer to the input samples
            data_in=insamples_ptr,
            # a pointer to the output buffer
            data_out=outsamples_ptr,
            # number of input samples
            input_frames=normalized_insamples.size,
            # number of output samples
            output_frames=outsamples.size,
            # NOT the end of input, i.e., there is more data to process
            end_of_input=int(end_of_input),
            # the conversion ratio
            src_ratio=self._ratio)
        sd_ptr = pointer(sd)
        rv = libsamplerate.src_process(self._state, sd_ptr)
        self._check_for_error(rv)
        return outsamples


class Resampler(Node):
    """
    `Resampler` expects to process :class:`~zounds.timeseries.AudioSamples`
    instances (e.g., those produced by a :class:`AudioStream` node), and will
    produce a new stream of :class:`AudioSamples` at a new sampling rate.

    Args:
        samplerate (AudioSampleRate): the desired sampling rate.  If none is
            provided, the default is :class:`~zounds.timeseries.SR44100`
        needs (Feature): a processing node that produces
            :class:`~zounds.timeseries.AudioSamples`


    Here's how you'd typically see :class:`Resampler` used in a processing
    graph.

    .. code:: python

        import featureflow as ff
        import zounds

        chunksize = zounds.ChunkSizeBytes(
            samplerate=zounds.SR44100(),
            duration=zounds.Seconds(30),
            bit_depth=16,
            channels=2)

        @zounds.simple_in_memory_settings
        class Document(ff.BaseModel):
            meta = ff.JSONFeature(
                zounds.MetaData,
                store=True,
                encoder=zounds.AudioMetaDataEncoder)

            raw = ff.ByteStreamFeature(
                ff.ByteStream,
                chunksize=chunksize,
                needs=meta,
                store=False)

            pcm = zounds.AudioSamplesFeature(
                zounds.AudioStream,
                needs=raw,
                store=True)

            resampled = zounds.AudioSamplesFeature(
                zounds.Resampler,
                samplerate=zounds.SR22050(),
                needs=pcm,
                store=True)


        synth = zounds.NoiseSynthesizer(zounds.SR11025())
        samples = synth.synthesize(zounds.Seconds(10))
        raw_bytes = samples.encode()
        _id = Document.process(meta=raw_bytes)
        doc = Document(_id)
        print doc.pcm.samplerate.__class__.__name__  # SR11025
        print doc.resampled.samplerate.__class__.__name__  # SR22050
    """

    def __init__(self, samplerate=None, needs=None):
        super(Resampler, self).__init__(needs=needs)
        self._samplerate = samplerate or SR44100()
        self._resample = None

    def _noop(self, data, finalized):
        return data

    def _process(self, data):
        sr = data.samples_per_second

        if self._resample is None:
            target_sr = self._samplerate.samples_per_second
            self._resample = Resample(
                sr,
                target_sr,
                1 if len(data.shape) == 1 else data.shape[1])

            if target_sr != sr:
                self._rs = self._resample
                # KLUDGE: The following line seems to solve a bug whereby 
                # libsamplerate doesn't generate enough samples the first time
                # src_process is called. We're calling it once here, so the "real"
                # output will come out click-free
                self._resample(np.zeros(target_sr, dtype=np.float32))
            else:
                self._rs = self._noop

        resampled = self._rs(data, self._finalized)
        if not isinstance(resampled, ArrayWithUnits):
            resampled = AudioSamples(resampled, self._samplerate)
        yield resampled



from __future__ import division
import numpy as np
import bisect


# TODO: What commonalities can be factored out of this class and TimeSlice?
# TODO: Deprecate stuff in psychoacoustics.py in favor of these classes
class FrequencyBand(object):
    """
    Represents an interval, or band of frequencies in hertz (cycles per second)

    Args:
        start_hz (float): The lower bound of the frequency band in hertz
        stop_hz (float): The upper bound of the frequency band in hertz

    Examples::
        >>> import zounds
        >>> band = zounds.FrequencyBand(500, 1000)
        >>> band.center_frequency
        750.0
        >>> band.bandwidth
        500
    """

    def __init__(self, start_hz, stop_hz):
        super(FrequencyBand, self).__init__()
        if stop_hz <= start_hz:
            raise ValueError('stop_hz must be greater than start_hz')
        self.stop_hz = stop_hz
        self.start_hz = start_hz

    def __eq__(self, other):
        try:
            return \
                self.start_hz == other.start_hz \
                and self.stop_hz == other.stop_hz
        except AttributeError:
            return super(FrequencyBand, self).__eq__(other)

    def __hash__(self):
        return (self.__class__.__name__, self.start_hz, self.stop_hz).__hash__()

    def intersect(self, other):
        """
        Return the intersection between this frequency band and another.

        Args:
            other (FrequencyBand): the instance to intersect with

        Examples::
            >>> import zounds
            >>> b1 = zounds.FrequencyBand(500, 1000)
            >>> b2 = zounds.FrequencyBand(900, 2000)
            >>> intersection = b1.intersect(b2)
            >>> intersection.start_hz, intersection.stop_hz
            (900, 1000)
        """
        lowest_stop = min(self.stop_hz, other.stop_hz)
        highest_start = max(self.start_hz, other.start_hz)
        return FrequencyBand(highest_start, lowest_stop)

    def bandwidth_ratio(self, other):
        return other.bandwidth / self.bandwidth

    def intersection_ratio(self, other):
        intersection = self.intersect(other)
        return self.bandwidth_ratio(intersection)

    @staticmethod
    def from_start(start_hz, bandwidth_hz):
        """
        Produce a :class:`FrequencyBand` instance from a lower bound and
        bandwidth

        Args:
            start_hz (float): the lower bound of the desired FrequencyBand
            bandwidth_hz (float): the bandwidth of the desired FrequencyBand

        """
        return FrequencyBand(start_hz, start_hz + bandwidth_hz)

    @staticmethod
    def from_center(center_hz, bandwidth_hz):
        half_bandwidth = bandwidth_hz / 2
        return FrequencyBand(
            center_hz - half_bandwidth, center_hz + half_bandwidth)

    @property
    def bandwidth(self):
        """
        The span of this frequency band, in hertz
        """
        return self.stop_hz - self.start_hz

    @property
    def center_frequency(self):
        return self.start_hz + (self.bandwidth / 2)

    def __repr__(self):
        return '''FrequencyBand(
start_hz={start_hz},
stop_hz={stop_hz},
center={center},
bandwidth={bandwidth})'''.format(
            start_hz=self.start_hz,
            stop_hz=self.stop_hz,
            center=self.center_frequency,
            bandwidth=self.bandwidth)


class FrequencyScale(object):
    """
    Represents a set of frequency bands with monotonically increasing start
    frequencies

    Args:
        frequency_band (FrequencyBand): A band representing the entire span of
            this scale.  E.g., one might want to generate a scale spanning the
            entire range of human hearing by starting with
            :code:`FrequencyBand(20, 20000)`
        n_bands (int): The number of bands in this scale
        always_even (bool): when converting frequency slices to integer indices
            that numpy can understand, should the slice size always be even?

    See Also:
        :class:`~zounds.spectral.LinearScale`
        :class:`~zounds.spectral.GeometricScale`
    """

    def __init__(self, frequency_band, n_bands, always_even=False):
        super(FrequencyScale, self).__init__()
        self.always_even = always_even
        self.n_bands = n_bands
        self.frequency_band = frequency_band
        self._bands = None
        self._starts = None
        self._stops = None

    @property
    def bands(self):
        """
        An iterable of all bands in this scale
        """
        if self._bands is None:
            self._bands = self._compute_bands()
        return self._bands

    @property
    def band_starts(self):
        if self._starts is None:
            self._starts = [b.start_hz for b in self.bands]
        return self._starts

    @property
    def band_stops(self):
        if self._stops is None:
            self._stops = [b.stop_hz for b in self.bands]
        return self._stops

    def _compute_bands(self):
        raise NotImplementedError()

    def __len__(self):
        return self.n_bands

    @property
    def center_frequencies(self):
        """
        An iterable of the center frequencies of each band in this scale
        """
        return (band.center_frequency for band in self)

    @property
    def bandwidths(self):
        """
        An iterable of the bandwidths of each band in this scale
        """
        return (band.bandwidth for band in self)

    def ensure_overlap_ratio(self, required_ratio=0.5):
        """
        Ensure that every adjacent pair of frequency bands meets the overlap
        ratio criteria.  This can be helpful in scenarios where a scale is
        being used in an invertible transform, and something like the `constant
        overlap add constraint
        <https://ccrma.stanford.edu/~jos/sasp/Constant_Overlap_Add_COLA_Cases.html>`_
        must be met in order to not introduce artifacts in the reconstruction.

        Args:
            required_ratio (float): The required overlap ratio between all
                adjacent frequency band pairs

        Raises:
            AssertionError: when the overlap ratio for one or more adjacent
                frequency band pairs is not met
        """

        msg = \
            'band {i}: ratio must be at least {required_ratio} but was {ratio}'

        for i in xrange(0, len(self) - 1):
            b1 = self[i]
            b2 = self[i + 1]

            try:
                ratio = b1.intersection_ratio(b2)
            except ValueError:
                ratio = 0

            if ratio < required_ratio:
                raise AssertionError(msg.format(**locals()))

    @property
    def Q(self):
        """
        The quality factor of the scale, or, the ratio of center frequencies
        to bandwidths
        """
        return np.array(list(self.center_frequencies)) \
               / np.array(list(self.bandwidths))

    @property
    def start_hz(self):
        """
        The lower bound of this frequency scale
        """
        return self.frequency_band.start_hz

    @property
    def stop_hz(self):
        """
        The upper bound of this frequency scale
        """
        return self.frequency_band.stop_hz

    def get_slice(self, frequency_band):
        """
        Given a frequency band, and a frequency dimension comprised of
        n_samples, return a slice using integer indices that may be used to
        extract only the frequency samples that intersect with the frequency
        band
        """
        start_index = bisect.bisect_left(
            self.band_stops, frequency_band.start_hz)
        stop_index = bisect.bisect_left(
            self.band_starts, frequency_band.stop_hz)

        if self.always_even and (stop_index - start_index) % 2:
            # KLUDGE: This is simple, but it may make sense to choose move the
            # upper *or* lower bound, based on which one introduces a lower
            # error
            stop_index += 1
        return slice(start_index, stop_index)

    def __eq__(self, other):
        return \
            self.__class__ == other.__class__ \
            and self.frequency_band == other.frequency_band \
            and self.n_bands == other.n_bands

    def __iter__(self):
        return iter(self.bands)

    def _construct_scale_from_slice(self, bands):
        freq_band = FrequencyBand(bands[0].start_hz, bands[-1].stop_hz)
        return self.__class__(freq_band, len(bands))

    def __getitem__(self, index):
        try:
            # index is an integer or slice
            bands = self.bands[index]
        except TypeError:
            # index is a frequency band
            bands = self.bands[self.get_slice(index)]

        if isinstance(bands, FrequencyBand):
            return bands

        return self._construct_scale_from_slice(bands)

    def __str__(self):
        cls = self.__class__.__name__
        return '{cls}(band={self.frequency_band}, n_bands={self.n_bands})' \
            .format(**locals())

    def __repr__(self):
        return self.__str__()


class LinearScale(FrequencyScale):
    """
    A linear frequency scale with constant bandwidth.  Appropriate for use
    with transforms whose coefficients also lie on a linear frequency scale,
    e.g. the FFT or DCT transforms.

    Args:
        frequency_band (FrequencyBand): A band representing the entire span of
            this scale.  E.g., one might want to generate a scale spanning the
            entire range of human hearing by starting with
            :code:`FrequencyBand(20, 20000)`
        n_bands (int): The number of bands in this scale
        always_even (bool): when converting frequency slices to integer indices
            that numpy can understand, should the slice size always be even?

    Examples:
        >>> from zounds import FrequencyBand, LinearScale
        >>> scale = LinearScale(FrequencyBand(20, 20000), 10)
        >>> scale
        LinearScale(band=FrequencyBand(
        start_hz=20,
        stop_hz=20000,
        center=10010.0,
        bandwidth=19980), n_bands=10)
        >>> scale.Q
        array([ 0.51001001,  1.51001001,  2.51001001,  3.51001001,  4.51001001,
                5.51001001,  6.51001001,  7.51001001,  8.51001001,  9.51001001])
    """

    def __init__(self, frequency_band, n_bands, always_even=False):
        super(LinearScale, self).__init__(frequency_band, n_bands, always_even)

    @staticmethod
    def from_sample_rate(sample_rate, n_bands, always_even=False):
        """
        Return a :class:`~zounds.spectral.LinearScale` instance whose upper
        frequency bound is informed by the nyquist frequency of the sample rate.

        Args:
            sample_rate (SamplingRate): the sample rate whose nyquist frequency
                will serve as the upper frequency bound of this scale
            n_bands (int): the number of evenly-spaced frequency bands
        """
        fb = FrequencyBand(0, sample_rate.nyquist)
        return LinearScale(fb, n_bands, always_even=always_even)

    def _compute_bands(self):
        freqs = np.linspace(
            self.start_hz, self.stop_hz, self.n_bands, endpoint=False)
        # constant, non-overlapping bandwidth
        bandwidth = freqs[1] - freqs[0]
        return tuple(FrequencyBand(f, f + bandwidth) for f in freqs)


class LogScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands, always_even=False):
        super(LogScale, self).__init__(
            frequency_band, n_bands, always_even=always_even)

    def _compute_bands(self):
        center_freqs = np.logspace(
            np.log10(self.start_hz),
            np.log10(self.stop_hz),
            self.n_bands + 1)
        # variable bandwidth
        bandwidths = np.diff(center_freqs)
        return tuple(FrequencyBand.from_center(cf, bw)
                     for (cf, bw) in zip(center_freqs[:-1], bandwidths))


class GeometricScale(FrequencyScale):
    """
    A constant-Q scale whose center frequencies progress geometrically rather
    than linearly

    Args:
        start_center_hz (int): the center frequency of the first band in the
            scale
        stop_center_hz (int): the center frequency of the last band in the scale
        bandwidth_ratio (float): the center frequency to bandwidth ratio
        n_bands (int): the total number of bands

    Examples:
        >>> from zounds import GeometricScale
        >>> scale = GeometricScale(20, 20000, 0.05, 10)
        >>> scale
        GeometricScale(band=FrequencyBand(
        start_hz=19.5,
        stop_hz=20500.0,
        center=10259.75,
        bandwidth=20480.5), n_bands=10)
        >>> scale.Q
        array([ 20.,  20.,  20.,  20.,  20.,  20.,  20.,  20.,  20.,  20.])
        >>> list(scale.center_frequencies)
        [20.000000000000004, 43.088693800637671, 92.831776672255558,
            200.00000000000003, 430.88693800637651, 928.31776672255558,
            2000.0000000000005, 4308.8693800637648, 9283.1776672255564,
            20000.000000000004]
    """

    def __init__(
            self,
            start_center_hz,
            stop_center_hz,
            bandwidth_ratio,
            n_bands,
            always_even=False):
        self.__bands = [
            FrequencyBand.from_center(cf, cf * bandwidth_ratio)
            for cf in np.geomspace(start_center_hz, stop_center_hz, num=n_bands)
            ]
        band = FrequencyBand(self.__bands[0].start_hz, self.__bands[-1].stop_hz)
        super(GeometricScale, self).__init__(
            band, n_bands, always_even=always_even)
        self.start_center_hz = start_center_hz
        self.stop_center_hz = stop_center_hz
        self.bandwidth_ratio = bandwidth_ratio

    def _construct_scale_from_slice(self, bands):
        return ExplicitScale(bands)

    def __eq__(self, other):
        return \
            super(GeometricScale, self).__eq__(other) \
            and self.start_center_hz == other.start_center_hz \
            and self.stop_center_hz == other.stop_center_hz \
            and self.bandwidth_ratio == other.bandwidth_ratio

    def _compute_bands(self):
        return self.__bands


class ExplicitScale(FrequencyScale):
    """
    A scale where the frequency bands are provided explicitly, rather than
    computed

    Args:
        bands (list of FrequencyBand): The explicit bands used by this scale

    See Also:
        :class:`~zounds.spectral.FrequencyAdaptive`
    """
    def __init__(self, bands):
        bands = list(bands)
        frequency_band = FrequencyBand(bands[0].start_hz, bands[-1].stop_hz)
        super(ExplicitScale, self).__init__(
            frequency_band, len(bands), always_even=False)
        self._bands = bands

    def _construct_scale_from_slice(self, bands):
        return ExplicitScale(bands)

    def _compute_bands(self):
        return self._bands

    def __eq__(self, other):
        return all([a == b for (a, b) in zip(self, other)])


class BarkScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands):
        super(BarkScale, self).__init__(frequency_band, n_bands)


class MelScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands):
        super(MelScale, self).__init__(frequency_band, n_bands)


class ChromaScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands):
        super(ChromaScale, self).__init__(frequency_band, n_bands)



import featureflow as ff


def simple_lmdb_settings(path, map_size=1e9, user_supplied_id=False):
    """
    Creates a decorator that can be used to configure sane default LMDB
    persistence settings for a model

    Args:
        path (str): The path where the LMDB database files will be created
        map_size (int): The amount of space to allot for the database
    """

    def decorator(cls):
        provider = \
            ff.UserSpecifiedIdProvider(key='_id') \
            if user_supplied_id else ff.UuidProvider()

        class Settings(ff.PersistenceSettings):
            id_provider = provider
            key_builder = ff.StringDelimitedKeyBuilder('|')
            database = ff.LmdbDatabase(
                    path, key_builder=key_builder, map_size=map_size)

        class Model(cls, Settings):
            pass

        Model.__name__ = cls.__name__
        Model.__module__ = cls.__module__
        return Model

    return decorator


def simple_object_storage_settings(container, region, username, api_key):
    def decorator(cls):
        class Settings(ff.PersistenceSettings):
            id_provider = ff.UuidProvider()
            key_builder = ff.StringDelimitedKeyBuilder()
            database = ff.ObjectStoreDatabase(
                container, username, api_key, region, key_builder=key_builder)

        class Model(cls, Settings):
            pass

        Model.__name__ = cls.__name__
        Model.__module__ = cls.__module__
        return Model

    return decorator


def simple_in_memory_settings(cls):
    """
    Decorator that returns a class that "persists" data in-memory.  Mostly
     useful for testing
    :param cls: the class whose features should be persisted in-memory
    :return: A new class that will persist features in memory
    """

    class Settings(ff.PersistenceSettings):
        id_provider = ff.UuidProvider()
        key_builder = ff.StringDelimitedKeyBuilder()
        database = ff.InMemoryDatabase(key_builder=key_builder)

    class Model(cls, Settings):
        pass

    Model.__name__ = cls.__name__
    Model.__module__ = cls.__module__
    return Model




import numpy as np
import unittest2
from functional import stft, apply_scale, frequency_decomposition
from zounds.core import ArrayWithUnits
from zounds.synthesize import SilenceSynthesizer, TickSynthesizer
from zounds.timeseries import SR22050, Milliseconds, TimeDimension, TimeSlice
from zounds.spectral import \
    HanningWindowingFunc, FrequencyDimension, LinearScale, GeometricScale, \
    ExplicitFrequencyDimension


class FrequencyDecompositionTests(unittest2.TestCase):
    def test_can_decompose(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(8192, 4096)
        duration = TimeSlice(wscheme.duration)
        frequency = TimeSlice(wscheme.frequency)
        _, windowed = samples.sliding_window_with_leftovers(
            duration, frequency, dopad=True)
        fa = frequency_decomposition(
            windowed, [32, 64, 128, 256, 512, 1024, 2048, 4096])
        self.assertEqual(windowed.dimensions[0], fa.dimensions[0])
        self.assertIsInstance(
            fa.dimensions[1], ExplicitFrequencyDimension)


class STFTTests(unittest2.TestCase):
    def test_has_correct_number_of_bins(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(6666))
        wscheme = sr.windowing_scheme(512, 256)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        self.assertEqual(tf.shape[1], 257)

    def test_has_correct_dimensions(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(6666))
        wscheme = sr.windowing_scheme(512, 256)
        tf = stft(samples, wscheme, HanningWindowingFunc())
        self.assertIsInstance(tf, ArrayWithUnits)
        self.assertEqual(2, len(tf.dimensions))
        self.assertIsInstance(tf.dimensions[0], TimeDimension)
        self.assertEqual(tf.dimensions[0].samplerate, wscheme)
        self.assertIsInstance(tf.dimensions[1], FrequencyDimension)
        self.assertIsInstance(tf.dimensions[1].scale, LinearScale)


class ApplyScaleTests(unittest2.TestCase):
    def test_has_correct_shape(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())
        self.assertEqual(tf.shape[:-1] + (len(scale),), geom.shape)

    def test_has_correct_dimensions(self):
        sr = SR22050()
        samples = SilenceSynthesizer(sr).synthesize(Milliseconds(9999))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())
        self.assertIsInstance(geom, ArrayWithUnits)
        self.assertEqual(2, len(geom.dimensions))
        self.assertIsInstance(geom.dimensions[0], TimeDimension)
        self.assertEqual(geom.dimensions[0].samplerate, wscheme)
        self.assertIsInstance(geom.dimensions[1], FrequencyDimension)
        self.assertEqual(scale, geom.dimensions[1].scale)

    def test_preserves_time_dimension(self):
        sr = SR22050()
        samples = TickSynthesizer(sr).synthesize(
            Milliseconds(10000), Milliseconds(1000))
        wscheme = sr.windowing_scheme(256, 128)
        scale = GeometricScale(50, sr.nyquist, 0.4, 32)
        scale.ensure_overlap_ratio()
        tf = stft(samples, wscheme, HanningWindowingFunc())
        geom = apply_scale(tf, scale, window=HanningWindowingFunc())

        # get the loudness envelope of each
        tf_envelope = np.abs(tf.real).sum(axis=1)
        geom_envelope = geom.sum(axis=1)

        tf_zeros = np.where(tf_envelope == 0)
        geom_zeros = np.where(geom_envelope == 0)

        np.testing.assert_allclose(tf_zeros, geom_zeros)



class Trainer(object):
    def __init__(self, epochs, batch_size):
        super(Trainer, self).__init__()
        self.batch_size = batch_size
        self.epochs = epochs

    def train(self):
        raise NotImplemented()



from trainer import Trainer
import numpy as np


class GanTrainer(Trainer):
    def __init__(
            self,
            network,
            loss,
            generator_optim_func,
            discriminator_optim_func,
            latent_dimension,
            epochs,
            batch_size):

        super(GanTrainer, self).__init__(epochs, batch_size)
        self.discriminator_optim_func = discriminator_optim_func
        self.generator_optim_func = generator_optim_func
        self.latent_dimension = latent_dimension
        self.loss = loss
        self.network = network
        self.discriminator = network.discriminator
        self.generator = network.generator

    def train(self, data):

        import torch
        from torch.autograd import Variable

        data = data.astype(np.float32)

        zdim = self.latent_dimension

        # TODO: These dimensions work for vanilla GANs, but need to be
        # reversed (batch_size, zdim, 1) for convolutional GANs

        noise_shape = (self.batch_size,) + self.latent_dimension
        noise = torch.FloatTensor(*noise_shape)
        fixed_noise = torch.FloatTensor(*noise_shape).normal_(0, 1)
        label = torch.FloatTensor(self.batch_size)
        real_label = 1
        fake_label = 0

        self.generator.cuda()
        self.discriminator.cuda()
        self.loss.cuda()
        label = label.cuda()
        noise, fixed_noise = noise.cuda(), fixed_noise.cuda()

        generator_optim = self.generator_optim_func(self.generator)
        discriminator_optim = self.discriminator_optim_func(self.discriminator)

        for epoch in xrange(self.epochs):
            for i in xrange(0, len(data), self.batch_size):
                minibatch = data[i: i + self.batch_size]

                if len(minibatch) != self.batch_size:
                    continue

                inp = torch.from_numpy(minibatch)
                inp = inp.cuda()

                # train discriminator on real data with one-sided label
                # smoothing
                self.discriminator.zero_grad()
                soft_labels = \
                    0.7 + (np.random.random_sample(self.batch_size) * 0.4) \
                        .astype(np.float32)
                soft_labels = torch.from_numpy(soft_labels)
                label.copy_(soft_labels)

                input_v = Variable(inp)
                label_v = Variable(label)

                output = self.discriminator.forward(input_v)
                real_error = self.loss(output, label_v)
                real_error.backward()

                # train discriminator on fake data
                noise.normal_(0, 1)
                noise_v = Variable(noise)
                fake = self.generator.forward(noise_v)
                label.fill_(fake_label)
                label_v = Variable(label)
                output = self.discriminator.forward(fake.detach())
                fake_error = self.loss(output, label_v)
                fake_error.backward()
                discriminator_error = real_error + fake_error
                discriminator_optim.step()

                # train generator
                self.generator.zero_grad()
                label.fill_(real_label)
                label_v = Variable(label)
                output = self.discriminator.forward(fake)
                generator_error = self.loss(output, label_v)
                generator_error.backward()
                generator_optim.step()

                gl = generator_error.data[0]
                dl = discriminator_error.data[0]
                re = real_error.data[0]
                fe = fake_error.data[0]

                if i % 10 == 0:
                    print \
                        'Epoch {epoch}, batch {i}, generator {gl}, real_error {re}, fake_error {fe}' \
                            .format(**locals())

        return self.network




class BaseDimensionEncoder(object):
    def __init__(self, dim_type):
        super(BaseDimensionEncoder, self).__init__()
        self.dim_type = dim_type

    def matches(self, o):
        return isinstance(o, self.dim_type)

    def dict(self, o):
        raise NotImplementedError()

    def encode(self, o):
        return dict(type=self.dim_type.__name__, data=self.dict(o))


class BaseDimensionDecoder(object):
    def __init__(self, dim_type):
        super(BaseDimensionDecoder, self).__init__()
        self.dim_type = dim_type

    def matches(self, d):
        return d['type'] == self.dim_type.__name__

    def args(self, d):
        return tuple()

    def kwargs(self, d):
        return dict()

    def decode(self, d):
        data = d['data']
        return self.dim_type(*self.args(data), **self.kwargs(data))



"""
The soundfile module introduces :class:`featureflow.Node` subclasses that know
how to process low-level audio samples and common audio encodings.
"""

from audio_metadata import MetaData, AudioMetaData, AudioMetaDataEncoder

from ogg_vorbis import \
    OggVorbis, OggVorbisDecoder, OggVorbisEncoder, OggVorbisFeature, \
    OggVorbisWrapper

from audiostream import AudioStream

from resample import Resampler

from chunksize import ChunkSizeBytes



__version__ = '0.25.12'

from timeseries import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds, \
    SR11025, SR16000, SR22050, SR44100, SR48000, SR96000, HalfLapped, Stride, \
    TimeSlice, VariableRateTimeSeries, VariableRateTimeSeriesFeature, \
    SampleRate, AudioSamples, TimeDimension, audio_sample_rate, \
    ConstantRateTimeSeries, nearest_audio_sample_rate

from soundfile import \
    MetaData, AudioMetaData, AudioMetaDataEncoder, OggVorbis, \
    OggVorbisDecoder, OggVorbisEncoder, OggVorbisFeature, OggVorbisWrapper, \
    AudioStream, Resampler, ChunkSizeBytes

from spectral import \
    SlidingWindow, OggVorbisWindowingFunc, WindowingFunc, \
    FFT, MDCT, DCT, DCTIV, BarkBands, Chroma, BFCC, SpectralCentroid, \
    SpectralFlatness, AWeighting, LinearScale, LogScale, FrequencyBand, \
    FrequencyScale, FrequencyDimension, GeometricScale, HanningWindowingFunc, \
    FrequencyAdaptiveTransform, ExplicitScale, ExplicitFrequencyDimension, \
    FrequencyAdaptive, FrequencyWeighting

from loudness import \
    log_modulus, inverse_log_modulus, decibel, mu_law, MuLaw, LogModulus, \
    inverse_mu_law

from segment import MeasureOfTransience, MovingAveragePeakPicker, \
    ComplexDomain, TimeSliceFeature

from synthesize import \
    FFTSynthesizer, DCTSynthesizer, TickSynthesizer, NoiseSynthesizer, \
    SineSynthesizer, DCTIVSynthesizer, MDCTSynthesizer, \
    FrequencyAdaptiveFFTSynthesizer, FrequencyAdaptiveDCTSynthesizer, \
    SilenceSynthesizer, WindowedAudioSynthesizer

from learn import \
    KMeans, Learned, MeanStdNormalization, UnitNorm, Log, Multiply, \
    PreprocessingPipeline, Slicer, ReservoirSampler, simple_settings, \
    SklearnModel, WithComponents, InstanceScaling, Reshape, ShuffledSamples, \
    PyTorchNetwork, PyTorchGan, PyTorchAutoEncoder, GanTrainer, \
    WassersteinGanTrainer, SupervisedTrainer, TripletEmbeddingTrainer, \
    Weighted, MuLawCompressed, SimHash, AbsoluteValue, Binarize, Sharpen, \
    learning_pipeline

from ui import ZoundsApp, ZoundsSearch, RangeUnitUnsupportedException

from index import \
    SearchResults, HammingDb, HammingIndex, BruteForceSearch, \
    HammingDistanceBruteForceSearch

from basic import \
    Slice, Sum, Max, Pooled, process_dir, stft, audio_graph, with_onsets, \
    resampled, frequency_adaptive

from util import \
    simple_lmdb_settings, simple_in_memory_settings, \
    simple_object_storage_settings

from nputil import sliding_window

from core import IdentityDimension, ArrayWithUnits

from persistence import \
    ArrayWithUnitsFeature, AudioSamplesFeature, FrequencyAdaptiveFeature

from datasets import \
    PhatDrumLoops, InternetArchive, FreeSoundSearch, DataSetCache, Directory, \
    ingest, MusicNet, NSynth


import sys
import json
import tornado.ioloop
import tornado.web
import httplib
import traceback
from cStringIO import StringIO
import uuid
from baseapp import BaseZoundsApp, NoMatchingSerializerException, RequestContext
from featureparser import FeatureParser


class ZoundsApp(BaseZoundsApp):
    """
    Adds an in-browser REPL to the base zounds application
    """
    def __init__(
            self,
            base_path=r'/zounds/',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            globals={},
            locals={}):

        super(ZoundsApp, self).__init__(
            base_path=base_path,
            model=model,
            visualization_feature=visualization_feature,
            audio_feature=audio_feature,
            html='index.html')

        self.globals = globals
        self.locals = locals
        self.temp = {}

    def custom_routes(self):
        return [
            (r'/zounds/temp/(.+?)/?', self.temp_handler()),
            (r'/zounds/repl/?', self.repl_handler())
        ]

    def temp_handler(self):

        app = self

        class TempHandler(tornado.web.RequestHandler):

            def get(self, _id):
                try:
                    result = app.temp[_id]
                except KeyError:
                    self.set_status(httplib.NOT_FOUND)
                    self.finish()
                    return
                self.set_header('Content-Type', result.content_type)
                self.set_header('Accept-Ranges', 'bytes')
                self.write(result.data)
                self.set_status(httplib.OK)
                self.finish()

        return TempHandler

    def repl_handler(self):
        document = self.model
        globals = self.globals
        locals = self.locals
        app = self

        class ReplHandler(tornado.web.RequestHandler):

            def _add_url(self, statement, output, value):
                parser = FeatureParser(document, locals)
                doc, feature = parser.parse_feature(statement)
                try:
                    context = RequestContext(
                            document=doc,
                            feature=feature,
                            value=value,
                            slce=slice(None))
                    result = app.serialize(context)
                    temp_id = uuid.uuid4().hex
                    app.temp[temp_id] = result
                    output['url'] = '/zounds/temp/{temp_id}'.format(
                            temp_id=temp_id)
                    output['contentType'] = result.content_type
                except NoMatchingSerializerException:
                    pass

            def post(self):
                statement = self.request.body
                self.set_header('Content-Type', 'application/json')
                output = dict()

                try:
                    try:
                        value = eval(statement, globals, locals)
                        output['result'] = str(value)
                        self._add_url(statement, output, value)
                    except SyntaxError:
                        orig_stdout = sys.stdout
                        sys.stdout = sio = StringIO()
                        exec (statement, globals, locals)
                        sio.seek(0)
                        output['result'] = sio.read()
                        sys.stdout = orig_stdout
                    self.set_status(httplib.OK)
                except:
                    output['error'] = traceback.format_exc()
                    self.set_status(httplib.BAD_REQUEST)

                self.write(json.dumps(output))
                self.finish()

        return ReplHandler



import unittest2
from chunksize import ChunkSizeBytes
from zounds.timeseries import SR44100, Seconds


class ChunkSizeBytesTests(unittest2.TestCase):
    def test_can_convert_to_integer_number_of_bytes(self):
        cs = ChunkSizeBytes(SR44100(), Seconds(30), channels=2, bit_depth=16)
        self.assertEqual(5292000, int(cs))

    def test_can_repr(self):
        cs = ChunkSizeBytes(SR44100(), Seconds(30), channels=2, bit_depth=16)
        s = cs.__repr__()
        self.assertEqual(
            'ChunkSizeBytes(samplerate=SR44100(f=2.2675736e-05, '
            'd=2.2675736e-05), duration=30 seconds, channels=2, bit_depth=16)',
            s)



import ast


class ExpressionVisitor(ast.NodeVisitor):
    def __init__(self, document, locals):
        super(ExpressionVisitor, self).__init__()
        self.locals = locals
        self.document = document
        self.feature_name = None
        self.doc = None

    @property
    def result(self):
        if self.feature_name:
            feature = self.document.features[self.feature_name]
        else:
            feature = None
        return self.doc, feature

    def visit_Expr(self, node):
        children = list(ast.iter_child_nodes(node))
        if len(children) != 1:
            raise ValueError()

        feature_name = None

        child = children[0]
        if isinstance(child, ast.Attribute) \
                and child.attr in self.document.features:
            feature_name = child.attr
        else:
            raise ValueError()

        grandchildren = list(ast.iter_child_nodes(child))
        if len(grandchildren) != 2:
            raise ValueError()

        grandchild = grandchildren[0]
        if isinstance(grandchild, ast.Name) \
                and grandchild.id in self.locals \
                and isinstance(self.locals[grandchild.id], self.document):
            self.doc = self.locals[grandchild.id]
            self.feature_name = feature_name
        else:
            raise ValueError()


class FeatureParser(object):
    def __init__(self, document, locals):
        super(FeatureParser, self).__init__()
        self.visitor = ExpressionVisitor(document, locals)

    def parse_feature(self, statement):
        root = ast.parse(statement)
        try:
            self.visitor.visit(root)
        except ValueError:
            pass
        return self.visitor.result



import unittest2
import numpy as np
from duration import Seconds
from samplerate import SR44100, SR11025, SampleRate
from zounds.timeseries import TimeDimension, TimeSlice
from zounds.core import IdentityDimension
from audiosamples import AudioSamples
from zounds.synthesize import SineSynthesizer


class AudioSamplesTest(unittest2.TestCase):

    def test_stereo(self):
        silence = AudioSamples.silence(SR11025(), Seconds(10))
        self.assertEqual(1, silence.channels)
        stereo = silence.stereo
        self.assertEqual(2, stereo.channels)
        self.assertEqual((len(silence), 2), stereo.shape)
        np.testing.assert_allclose(silence, stereo[:, 0])
        np.testing.assert_allclose(silence, stereo[:, 1])

    def test_silence_creates_silence(self):
        silence = AudioSamples.silence(SR11025(), Seconds(10))
        self.assertEqual(0, silence.sum())

    def test_pad_with_samples_adds_silence_at_end(self):
        synth = SineSynthesizer(SR11025())
        samples = synth.synthesize(Seconds(2))
        padded = samples.pad_with_silence(Seconds(4))
        silence = padded[TimeSlice(start=Seconds(2))]
        self.assertEqual(0, silence.sum())

    def test_raises_if_not_audio_samplerate(self):
        arr = np.zeros(int(44100 * 2.5))
        one = Seconds(1)
        self.assertRaises(
            TypeError, lambda: AudioSamples(arr, SampleRate(one, one)))

    def test_raises_if_array_is_more_than_2d(self):
        arr = np.zeros((int(44100 * 2.5), 2, 2))
        self.assertRaises(
            ValueError, lambda: AudioSamples(arr, SR44100()))

    def test_can_create_instance(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance, AudioSamples)
        length_seconds = instance.end / Seconds(1)
        self.assertAlmostEqual(2.5, length_seconds, places=6)

    def test_can_mix_two_instances(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = AudioSamples(arr, SR44100())
        mixed = first + second
        self.assertIsInstance(mixed, AudioSamples)
        self.assertEqual(SR44100(), mixed.samplerate)
        np.testing.assert_allclose(mixed, 2)

    def test_cannot_mix_two_instances_with_different_sample_rates(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = AudioSamples(arr, SR11025())
        self.assertRaises(ValueError, lambda: first + second)

    def test_can_add_plain_numpy_array(self):
        arr = np.ones(int(44100 * 2.5))
        first = AudioSamples(arr, SR44100())
        second = arr.copy()
        mixed = first + second
        self.assertIsInstance(mixed, AudioSamples)
        self.assertEqual(SR44100(), mixed.samplerate)
        np.testing.assert_allclose(mixed, 2)

    def test_channels_returns_one_for_one_dimensional_array(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(1, instance.channels)

    def test_channels_returns_two_for_two_dimensional_array(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        self.assertEqual(2, instance.channels)

    def test_samplerate_returns_correct_value(self):
        arr = np.zeros(int(44100 * 2.5))
        instance = AudioSamples(arr, SR44100())
        self.assertIsInstance(instance.samplerate, SR44100)

    def test_can_sum_to_mono(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        instance = AudioSamples(arr, SR44100())
        mono = instance.mono
        self.assertEqual(1, mono.channels)
        self.assertIsInstance(mono.samplerate, SR44100)

    def test_class_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = AudioSamples.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_instance_concat_returns_audio_samples(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR44100())
        s3 = s1.concat([s1, s2])
        self.assertIsInstance(s3, AudioSamples)
        self.assertEqual(44100 * 3, len(s3))

    def test_concat_raises_for_different_sample_rates(self):
        s1 = AudioSamples(np.zeros(44100 * 2), SR44100())
        s2 = AudioSamples(np.zeros(44100), SR11025())
        self.assertRaises(ValueError, lambda: AudioSamples.concat([s1, s2]))

    def test_sum_along_time_axis(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        ts = AudioSamples(arr, SR44100())
        result = ts.sum(axis=0)
        self.assertIsInstance(result, np.ndarray)
        self.assertNotIsInstance(result, AudioSamples)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)

    def test_sum_along_second_axis(self):
        arr = np.zeros(int(44100 * 2.5))
        arr = np.column_stack((arr, arr))
        ts = AudioSamples(arr, SR44100())
        result = ts.sum(axis=1)
        self.assertIsInstance(result, AudioSamples)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], TimeDimension)



import unittest2
from zounds.core import ArrayWithUnits
from zounds.timeseries import TimeDimension, Seconds, Milliseconds, SR11025
from zounds.spectral import \
    FrequencyBand, ExplicitFrequencyDimension, GeometricScale, ExplicitScale, \
    FrequencyDimension, LinearScale
from frequencyadaptive import FrequencyAdaptive
import numpy as np


class FrequencyAdaptiveTests(unittest2.TestCase):
    def test_raises_when_time_dimension_is_none(self):
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        self.assertRaises(
            ValueError, lambda: FrequencyAdaptive(arrs, None, scale))

    def test_raises_when_explicit_freq_dimension_and_non_contiguous_array(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa1 = FrequencyAdaptive(arrs, td, scale)
        arr = np.asarray(fa1)
        self.assertRaises(ValueError, lambda: FrequencyAdaptive(
            arr,
            td,
            scale=scale,
            explicit_freq_dimension=fa1.frequency_dimension))

    def test_can_construct_from_contiguous_array(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa1 = FrequencyAdaptive(arrs, td, scale)
        arr = np.asarray(fa1)
        fa2 = FrequencyAdaptive(
            arr, td, explicit_freq_dimension=fa1.frequency_dimension)
        self.assertEqual(fa1.shape, fa2.shape)
        self.assertEqual(fa1.scale, fa2.scale)
        self.assertEqual(fa1.time_dimension, fa2.time_dimension)
        self.assertEqual(fa1.frequency_dimension, fa2.frequency_dimension)

    def test_can_construct_instance(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)
        self.assertEqual((10, 55), fa.shape)
        self.assertIsInstance(fa.dimensions[0], TimeDimension)
        self.assertIsInstance(fa.dimensions[1], ExplicitFrequencyDimension)

    def test_can_concatenate_instances(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)

        arrs2 = [np.zeros((20, x)) for x in xrange(1, 11)]
        fa2 = FrequencyAdaptive(arrs2, td, scale)

        result = fa.concatenate(fa2)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((30, 55), result.shape)

    def test_can_get_single_frequency_band_over_entire_duration(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)

        single_band = fa[:, scale[5]]

        self.assertIsInstance(single_band, ArrayWithUnits)
        self.assertIsInstance(single_band.dimensions[0], TimeDimension)
        self.assertIsInstance(
            single_band.dimensions[1], ExplicitFrequencyDimension)
        self.assertEqual(1, len(single_band.dimensions[1].scale))
        self.assertEqual(1, len(single_band.dimensions[1].slices))

    def test_can_apply_frequency_slice_across_multiple_bands(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)
        band = FrequencyBand(300, 3030)

        fa2 = fa[:, band]

        self.assertIsInstance(fa2, ArrayWithUnits)
        self.assertEqual(td, fa2.dimensions[0])
        self.assertIsInstance(fa2.dimensions[1], ExplicitFrequencyDimension)
        self.assertIsInstance(fa2.dimensions[1].scale, ExplicitScale)

    def test_can_assign_to_multi_band_frequency_slice(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)
        band = FrequencyBand(300, 3030)
        fa[:, band] = 1
        int_slice = fa.dimensions[1].integer_based_slice(band)
        np.testing.assert_allclose(fa[:, int_slice], 1)

    def test_can_access_single_frequency_band(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.zeros((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        sliced = fa[:, scale[0]]
        self.assertEqual((10, 1), sliced.shape)

    def test_rasterize_does_not_distort_spectral_shape(self):
        n_bands = 8
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(
            start_center_hz=50,
            stop_center_hz=5000,
            bandwidth_ratio=0.07123,
            n_bands=n_bands)
        arrs = [np.random.normal(0, 1, (10, 2 ** (i + 1))) for i in
                xrange(n_bands)]
        maxes = [arr.max() for arr in arrs]
        fa = FrequencyAdaptive(arrs, td, scale)
        rasterized = fa.rasterize(n_coeffs=32)
        np.testing.assert_allclose(rasterized.max(axis=(0, 1)), maxes)

    def test_rasterize_gracefully_handles_band_with_no_energy(self):
        n_bands = 8
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(
            start_center_hz=50,
            stop_center_hz=5000,
            bandwidth_ratio=0.07123,
            n_bands=n_bands)
        arrs = [np.random.normal(0, 1, (10, 2 ** (i + 1))) for i in
                xrange(n_bands)]
        arrs[0][:] = 0
        fa = FrequencyAdaptive(arrs, td, scale)
        rasterized = fa.rasterize(n_coeffs=32)
        np.testing.assert_allclose(rasterized[:, :,  0], 0)

    def test_square_form_with_overlap(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.zeros((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        square = fa.square(50)

        self.assertEqual(3, square.ndim)
        self.assertEqual(10, square.shape[0])
        self.assertEqual(50, square.shape[1])
        self.assertEqual(120, square.shape[2])

        self.assertIsInstance(square, ArrayWithUnits)

        self.assertIsInstance(square.dimensions[0], TimeDimension)
        self.assertEqual(Milliseconds(5500), square.dimensions[0].end)
        self.assertEqual(Milliseconds(500), square.dimensions[0].frequency)
        self.assertEqual(Milliseconds(1000), square.dimensions[0].duration)

        self.assertIsInstance(square.dimensions[1], TimeDimension)

        self.assertIsInstance(square.dimensions[2], FrequencyDimension)
        self.assertEqual(scale, square.dimensions[2].scale)

    def test_square_form_with_overlap_do_overlap_add(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Milliseconds(500))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.zeros((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        square = fa.square(50, do_overlap_add=True)

        self.assertEqual(2, square.ndim)
        self.assertEqual(275, square.shape[0])
        self.assertEqual(120, square.shape[1])

        self.assertIsInstance(square, ArrayWithUnits)

        self.assertIsInstance(square.dimensions[0], TimeDimension)
        self.assertEqual(Milliseconds(5500), square.dimensions[0].end)
        self.assertEqual(Milliseconds(20), square.dimensions[0].frequency)
        self.assertEqual(Milliseconds(20), square.dimensions[0].duration)

        self.assertIsInstance(square.dimensions[1], FrequencyDimension)
        self.assertEqual(scale, square.dimensions[1].scale)

    def test_square_form_no_overlap(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.zeros((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        square = fa.square(50)

        self.assertEqual(3, square.ndim)

        self.assertEqual(10, square.shape[0])
        self.assertEqual(50, square.shape[1])
        self.assertEqual(120, square.shape[2])

        self.assertIsInstance(square, ArrayWithUnits)

        self.assertIsInstance(square.dimensions[0], TimeDimension)
        self.assertEqual(Seconds(10), square.dimensions[0].end)
        self.assertEqual(Milliseconds(1000), square.dimensions[0].frequency)
        self.assertEqual(Milliseconds(1000), square.dimensions[0].duration)

        self.assertIsInstance(square.dimensions[1], TimeDimension)

        self.assertIsInstance(square.dimensions[2], FrequencyDimension)
        self.assertEqual(scale, square.dimensions[2].scale)

    def test_square_form_no_overlap_do_overlap_add(self):
        td = TimeDimension(
            duration=Seconds(1),
            frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 120)
        arrs = [np.zeros((10, x)) for x in xrange(1, 121)]
        fa = FrequencyAdaptive(arrs, td, scale)
        square = fa.square(50, do_overlap_add=True)

        self.assertEqual(2, square.ndim)
        self.assertEqual(500, square.shape[0])
        self.assertEqual(120, square.shape[1])

        self.assertIsInstance(square, ArrayWithUnits)

        self.assertIsInstance(square.dimensions[0], TimeDimension)
        self.assertEqual(Seconds(10), square.dimensions[0].end)
        self.assertEqual(Milliseconds(20), square.dimensions[0].frequency)
        self.assertEqual(Milliseconds(20), square.dimensions[0].duration)

        self.assertIsInstance(square.dimensions[1], FrequencyDimension)
        self.assertEqual(scale, square.dimensions[1].scale)

    def test_from_arr_with_units(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)

        raw_arr = ArrayWithUnits(np.array(fa), fa.dimensions)

        fa2 = FrequencyAdaptive.from_array_with_units(raw_arr)
        self.assertIsInstance(fa2, FrequencyAdaptive)
        self.assertEqual(fa2.dimensions[0], fa.dimensions[0])
        self.assertEqual(fa2.dimensions[1], fa.dimensions[1])
        self.assertEqual(fa.shape, fa2.shape)




class ChunkSizeBytes(object):
    """
    A convenience class to help describe a chunksize in bytes for the
    :class:`featureflow.ByteStream` in terms of audio sample batch sizes.

    Args:
        samplerate (SampleRate): The samples-per-second factor
        duration (numpy.timedelta64): The length of desired chunks in seconds
        channels (int): Then audio channels factor
        bit_depth (int): The bit depth factor

    Examples:
        >>> from zounds import ChunkSizeBytes, Seconds, SR44100
        >>> chunksize = ChunkSizeBytes(SR44100(), Seconds(30))
        >>> chunksize
        ChunkSizeBytes(samplerate=SR44100(f=2.2675736e-05, d=2.2675736e-05)...
        >>> int(chunksize)
        5292000

    """
    def __init__(self, samplerate, duration, channels=2, bit_depth=16):
        self.duration = duration
        self.bit_depth = bit_depth
        self.channels = channels
        self.samplerate = samplerate

    def __int__(self):
        byte_depth = self.bit_depth // 8
        total_samples = int(self.duration / self.samplerate.frequency)
        return int(total_samples * byte_depth * self.channels)

    def __repr__(self):
        msg = 'ChunkSizeBytes(samplerate={samplerate}, duration={duration}, ' \
              'channels={channels}, bit_depth={bit_depth})'

        return msg.format(
            samplerate=self.samplerate,
            duration=str(self.duration),
            channels=self.channels,
            bit_depth=self.bit_depth)



import numpy as np
from zounds.loudness import mu_law, inverse_mu_law
from zounds.core import ArrayWithUnits, IdentityDimension


def categorical(x, mu=255):
    """
    Mu-law compress a block of audio samples, and convert them into a
    categorical distribution
    """
    # normalize the signal
    mx = x.max()
    x = np.divide(x, mx, where=mx != 0)

    # mu law compression
    x = mu_law(x)

    # translate and scale to [0, 1]
    x = (x - x.min()) * 0.5

    # convert to the range [0, 255]
    x = (x * mu).astype(np.uint8)

    # create the array to house the categorical representation
    c = np.zeros((np.product(x.shape), mu + 1), dtype=np.uint8)
    c[np.arange(len(c)), x.flatten()] = 1

    return ArrayWithUnits(
        c.reshape(x.shape + (mu + 1,)),
        x.dimensions + (IdentityDimension(),))


def inverse_categorical(x, mu=255):
    """
    Invert categorical samples
    """
    flat = x.reshape((-1, x.shape[-1]))
    indices = np.argmax(flat, axis=1).astype(np.float32)
    indices = (indices / mu) - 0.5
    inverted = inverse_mu_law(indices, mu=mu).reshape(x.shape[:-1])
    return ArrayWithUnits(inverted, x.dimensions[:2])




from io import BytesIO


class PreDownload(BytesIO):
    def __init__(self, initial_bytes, url):
        super(PreDownload, self).__init__(initial_bytes)
        if not url:
            raise ValueError('url must be provided')
        self.url = url



import unittest2
from index import HammingIndex
from featureflow import \
    PersistenceSettings, UuidProvider, StringDelimitedKeyBuilder, \
    InMemoryDatabase, InMemoryChannel, EventLog
from zounds.basic import stft, Slice, Binarize
from zounds.timeseries import SR11025, Seconds
from zounds.synthesize import SineSynthesizer
from zounds.persistence import ArrayWithUnitsFeature
from zounds.soundfile import AudioMetaData
import shutil
from uuid import uuid4


class HammingIndexTests(unittest2.TestCase):
    def setUp(self):
        self.event_log_path = '/tmp/{path}'.format(path=uuid4().hex)
        self.hamming_db_path = '/tmp/{path}'.format(path=uuid4().hex)

    def tearDown(self):
        shutil.rmtree(self.event_log_path, ignore_errors=True)
        shutil.rmtree(self.hamming_db_path, ignore_errors=True)

    def test_listen_raises_if_model_class_has_no_event_log_configured(self):
        Model = self._model(
            slice_size=64,
            settings=self._settings_with_no_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        self.assertRaises(
            ValueError, lambda: index._synchronously_process_events())

    def test_correctly_infers_code_size_8(self):
        Model = self._model(
            slice_size=64,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()
        self.assertEqual(8, index.hamming_db.code_size)

    def test_hamming_db_is_initialized_if_docs_exist(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()

        index2 = self._index(Model, Model.sliced)
        self.assertIsNotNone(index2.hamming_db)
        self.assertEqual(16, index2.hamming_db.code_size)

    def test_correctly_infers_code_size_16(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()
        self.assertEqual(16, index.hamming_db.code_size)

    def test_can_roundtrip_query(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()

        results = index.random_search(n_results=5)
        decoded = index.decode_query(results.query)
        encoded = index.encode_query(decoded)
        self.assertEqual(results.query, encoded)

    def test_can_search_with_binary_code(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()

        results = index.random_search(n_results=5)
        decoded = index.decode_query(results.query)
        encoded = index.encode_query(decoded)
        results = index.search(encoded, 5)
        self.assertEqual(5, len(list(results)))

    def test_can_add_additional_data_to_index(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(
            Model,
            Model.sliced,
            web_url=lambda doc, ts: doc.meta['web_url'])

        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        meta = AudioMetaData(uri=signal.encode(), web_url='https://example.com')
        _id = Model.process(meta=meta)
        index._synchronously_process_events()

        results = list(index.random_search(n_results=5))
        result_id, ts, extra_data = results[0]
        self.assertEqual(_id, result_id)
        self.assertEqual('https://example.com', extra_data['web_url'])

    def correctly_infers_index_name(self):
        Model = self._model(
            slice_size=128,
            settings=self._settings_with_event_log())

        index = self._index(Model, Model.sliced)
        signal = SineSynthesizer(SR11025()) \
            .synthesize(Seconds(5), [220, 440, 880])
        Model.process(meta=signal.encode())
        index._synchronously_process_events()
        self.assertTrue('index.sliced' in index.hamming_db.path)

    def _settings_with_no_event_log(self):
        class Settings(PersistenceSettings):
            id_provider = UuidProvider()
            key_builder = StringDelimitedKeyBuilder()
            database = InMemoryDatabase(key_builder=key_builder)

        return Settings

    def _settings_with_event_log(self):
        class Settings(PersistenceSettings):
            id_provider = UuidProvider()
            key_builder = StringDelimitedKeyBuilder()
            database = InMemoryDatabase(key_builder=key_builder)
            event_log = EventLog(
                path=self.event_log_path,
                channel=InMemoryChannel())

        return Settings

    def _model(self, slice_size, settings):
        STFT = stft(resample_to=SR11025(), store_fft=True)

        class Model(STFT, settings):
            binary = ArrayWithUnitsFeature(
                Binarize,
                predicate=lambda data: data >= 0,
                needs=STFT.fft,
                store=False)

            sliced = ArrayWithUnitsFeature(
                Slice,
                sl=slice(0, slice_size),
                needs=binary,
                store=True)

        return Model

    def _index(self, document, feature, **extra_data):
        return HammingIndex(
            document,
            feature,
            path=self.hamming_db_path,
            **extra_data)



from trainer import Trainer
import numpy as np

class WassersteinGanTrainer(Trainer):
    """
    Args:
        preprocess (function): function that takes the current epoch, and
        a minibatch, and mutates the minibatch
        arg_maker (function): function that takes the current epoch and outputs
            args to pass to the generator and discriminator
    """

    def __init__(
            self,
            network,
            latent_dimension,
            n_critic_iterations,
            epochs,
            batch_size,
            preprocess=None,
            arg_maker=None,
            generator_loss_term=lambda network, output: 0,
            critic_loss_term=lambda network, output: 0):

        super(WassersteinGanTrainer, self).__init__(epochs, batch_size)
        self.critic_loss_term = critic_loss_term
        self.generator_loss_term = generator_loss_term
        self.arg_maker = arg_maker
        self.preprocess = preprocess
        self.n_critic_iterations = n_critic_iterations
        self.latent_dimension = latent_dimension
        self.network = network
        self.critic = network.discriminator
        self.generator = network.generator

    def _minibatch(self, data):
        indices = np.random.randint(0, len(data), self.batch_size)
        return data[indices, ...]

    def _gradient_penalty(self, real_samples, fake_samples):
        """
        Compute the norm of the gradients for each sample in a batch, and
        penalize anything on either side of unit norm
        """
        import torch
        from torch.autograd import Variable, grad

        # computing the norm of the gradients is very expensive, so I'm only
        # taking a subset of the minibatch here
        subset_size = 10

        real_samples = real_samples[:subset_size]
        fake_samples = fake_samples[:subset_size]

        # TODO: this should have the same number of dimensions as real and
        # fake samples, and should not be hard-coded
        alpha = torch.rand(subset_size, 1).cuda()

        interpolates = alpha * real_samples + ((1 - alpha) * fake_samples)
        interpolates = Variable(interpolates.cuda(), requires_grad=True)

        d_output = self.critic(interpolates)

        gradients = grad(
            outputs=d_output,
            inputs=interpolates,
            grad_outputs=torch.ones(d_output.size()).cuda(),
            create_graph=True,
            retain_graph=True,
            only_inputs=True)[0]
        return ((gradients.norm(2, dim=1) - 1) ** 2).mean() * 10

    def train(self, data):

        import torch
        from torch.optim import Adam
        from torch.autograd import Variable

        data = data.astype(np.float32)

        zdim = self.latent_dimension

        noise_shape = (self.batch_size,) + self.latent_dimension
        noise = torch.FloatTensor(*noise_shape)
        fixed_noise = torch.FloatTensor(*noise_shape).normal_(0, 1)

        self.generator.cuda()
        self.critic.cuda()
        noise, fixed_noise = noise.cuda(), fixed_noise.cuda()

        trainable_generator_params = (
            p for p in self.generator.parameters() if p.requires_grad)
        trainable_critic_params = (
            p for p in self.critic.parameters() if p.requires_grad)

        generator_optim = Adam(
            trainable_generator_params, lr=0.0001, betas=(0, 0.9))
        critic_optim = Adam(
            trainable_critic_params, lr=0.0001, betas=(0, 0.9))

        for epoch in xrange(self.epochs):
            if self.arg_maker:
                kwargs = self.arg_maker(epoch)
            else:
                kwargs = dict()

            for i in xrange(0, len(data), self.batch_size):
                self.generator.zero_grad()
                self.critic.zero_grad()

                self.generator.eval()
                self.critic.train()

                for c in xrange(self.n_critic_iterations):

                    self.critic.zero_grad()

                    minibatch = self._minibatch(data)
                    inp = torch.from_numpy(minibatch)
                    inp = inp.cuda()
                    input_v = Variable(inp)

                    if self.preprocess:
                        input_v = self.preprocess(epoch, input_v)

                    d_real = self.critic.forward(input_v, **kwargs)

                    # train discriminator on fake data
                    noise.normal_(0, 1)
                    noise_v = Variable(noise, volatile=True)
                    fake = Variable(
                        self.generator.forward(noise_v, **kwargs).data)

                    if self.preprocess:
                        fake = self.preprocess(epoch, fake)

                    d_fake = self.critic.forward(fake, **kwargs)

                    real_mean = torch.mean(d_real)
                    fake_mean = torch.mean(d_fake)
                    d_loss = \
                        (fake_mean - real_mean) \
                        + self._gradient_penalty(input_v.data, fake.data) \
                        + self.critic_loss_term(self.critic, d_fake)
                    d_loss.backward()
                    critic_optim.step()

                self.generator.train()
                self.critic.eval()

                # train generator
                noise.normal_(0, 1)
                noise_v = Variable(noise)
                fake = self.generator.forward(noise_v, **kwargs)

                if self.preprocess:
                    fake = self.preprocess(epoch, fake)

                d_fake = self.critic.forward(fake, **kwargs)
                g_loss = \
                    -torch.mean(d_fake) \
                    + self.generator_loss_term(self.generator, fake)
                g_loss.backward()
                generator_optim.step()

                gl = g_loss.data[0]
                dl = d_loss.data[0]

                if i % 10 == 0:
                    print \
                        'Epoch {epoch}, batch {i}, generator {gl}, critic {dl}' \
                            .format(**locals())

        return self.network



import unittest2
from hammingdb import HammingDb
from uuid import uuid4
import shutil
import numpy as np
import os


class HammingDbTests(unittest2.TestCase):

    def setUp(self):
        self._db_name = uuid4().hex
        self._path = '/tmp/{path}'.format(path=self._db_name)

        def extract_code_from_text(text, n_chunks=1):
            n_bits = n_chunks * 64
            code = np.zeros(n_bits, dtype=np.bool)
            for i in xrange(len(text)):
                trigram = text[i: i + 3]
                hashed = hash(trigram)
                code[hashed % n_bits] = 1
            return np.packbits(code).tostring()
        self.extract_code_from_text = extract_code_from_text

    def tearDown(self):
        try:
            shutil.rmtree(self._path)
        except OSError:
            pass

    def test_can_initialize_with_no_code_size_when_data_already_exists(self):
        HammingDb(self._path, code_size=32)
        db2 = HammingDb(self._path, code_size=None)
        self.assertEqual(32, db2.code_size)

    def test_raises_when_persisted_code_size_does_not_agree_with_init(self):
        HammingDb(self._path, code_size=32)
        self.assertRaises(
            ValueError, lambda: HammingDb(self._path, code_size=16))

    def test_raises_when_code_size_is_none_for_uninitialized_database(self):
        self.assertRaises(
            ValueError, lambda: HammingDb(self._path, code_size=None))

    def test_metadata_fetch_does_not_raise_when_unitialized(self):
        db = HammingDb(self._path, code_size=16)
        v = db.get_metadata('cat')
        self.assertEqual(None, v)

    def test_can_set_and_get_metadata(self):
        db = HammingDb(self._path, code_size=16)
        db.set_metadata('cat', 'dog')
        self.assertEqual('dog', db.get_metadata('cat'))

    def test_can_get_random_entry(self):
        db = HammingDb(self._path, code_size=16)
        for i in xrange(100):
            db.append(os.urandom(16), str(i))
        code, results = db.random_search(10)
        results = list(results)
        self.assertEqual(10, len(results))

    def test_can_create_database_with_128_bit_codes(self):
        db = HammingDb(self._path, code_size=16)
        self.assertEqual(0, len(db))

    def test_can_append_with_128_bits(self):
        db = HammingDb(self._path, code_size=16)
        db.append('a' * 16, 'some data')
        self.assertEqual(1, len(db))

    def test_can_search_with_128_bits(self):
        db = HammingDb(self._path, code_size=16)
        t1 = 'Mary had a little lamb'
        t2 = 'Mary had a little dog'
        t3 = 'Permanent Midnight'
        t4 = 'Mary sad a little cog'
        extract_code = lambda x: self.extract_code_from_text(x, n_chunks=2)
        db.append(extract_code(t1), t1)
        db.append(extract_code(t2), t2)
        db.append(extract_code(t3), t3)
        db.append(extract_code(t4), t4)
        results = list(db.search(extract_code(t1), 3))
        self.assertEqual(3, len(results))
        self.assertEqual(t1, results[0])
        self.assertEqual(t2, results[1])
        self.assertEqual(t4, results[2])

    def test_constructor_raises_when_not_multiple_of_eight(self):
        self.assertRaises(
            ValueError, lambda: HammingDb(self._path, code_size=7))

    def test_empty_db_has_zero_length(self):
        db = HammingDb(self._path, code_size=8)
        self.assertEqual(0, len(db))

    def test_db_has_length_one_after_appending(self):
        db = HammingDb(self._path, code_size=8)
        db.append('a' * 8, 'some data')
        self.assertEqual(1, len(db))

    def test_db_has_length_two_after_appending_twice(self):
        db = HammingDb(self._path, code_size=8)
        db.append('a' * 8, 'some data')
        db.append('a' * 8, 'some data')
        self.assertEqual(2, len(db))

    def test_can_search_over_text_documents(self):
        db = HammingDb(self._path, code_size=8)
        t1 = 'Mary had a little lamb'
        t2 = 'Mary had a little dog'
        t3 = 'Permanent Midnight'
        t4 = 'Mary sad a little cog'
        db.append(self.extract_code_from_text(t1), t1)
        db.append(self.extract_code_from_text(t2), t2)
        db.append(self.extract_code_from_text(t3), t3)
        db.append(self.extract_code_from_text(t4), t4)
        results = list(db.search(self.extract_code_from_text(t1), 3))
        self.assertEqual(3, len(results))
        self.assertEqual(t1, results[0])
        self.assertEqual(t2, results[1])
        self.assertEqual(t4, results[2])

    def test_can_search_over_data_added_from_another_instance(self):
        db = HammingDb(self._path, code_size=8)
        db2 = HammingDb(self._path, code_size=8)
        t1 = 'Mary had a little lamb'
        t2 = 'Mary had a little dog'
        t3 = 'Permanent Midnight'
        t4 = 'Mary sad a little cog'
        db.append(self.extract_code_from_text(t1), t1)
        db.append(self.extract_code_from_text(t2), t2)
        db.append(self.extract_code_from_text(t3), t3)
        db.append(self.extract_code_from_text(t4), t4)
        results = list(db2.search(self.extract_code_from_text(t1), 3))
        self.assertEqual(3, len(results))
        s = set(results)
        self.assertTrue(t1 in s)
        self.assertTrue(t2 in s)
        self.assertTrue(t4 in s)

    def test_cannot_append_wrong_code_size(self):
        db = HammingDb(self._path, code_size=8)
        self.assertRaises(ValueError, lambda: db.append('a' * 7, 'some data'))

    def test_cannot_search_for_wrong_code_size(self):
        db = HammingDb(self._path, code_size=8)
        self.assertRaises(ValueError, lambda: list(db.search('a' * 7, 10)))

    def test_external_modifications_are_detected(self):
        db = HammingDb(self._path, code_size=8)
        db2 = HammingDb(self._path, code_size=8)
        db2.append('a' * 8, 'some data')
        self.assertEqual(1, len(db))

    def test_external_modifications_are_detected_when_db_has_keys(self):
        db = HammingDb(self._path, code_size=8)
        db2 = HammingDb(self._path, code_size=8)
        db.append('a' * 8, 'some other data')
        db2.append('a' * 8, 'some data')
        self.assertEqual(2, len(db))
        self.assertEqual(2, len(db2))

    def test_db_starts_with_correct_number_of_keys(self):
        db2 = HammingDb(self._path, code_size=8)
        db2.append('a' * 8, 'some data')
        db = HammingDb(self._path, code_size=8)
        self.assertEqual(1, len(db))

    def test_can_retrieve_data_from_search(self):
        db = HammingDb(self._path, code_size=8)
        t1 = 'Mary had a little lamb'
        t2 = 'Mary had a little dog'
        t3 = 'Permanent Midnight'
        t4 = 'Mary sad a little cog'
        db.append(self.extract_code_from_text(t1), t1)
        db.append(self.extract_code_from_text(t2), t2)
        db.append(self.extract_code_from_text(t3), t3)
        db.append(self.extract_code_from_text(t4), t4)
        results = list(db.search(self.extract_code_from_text(t1), 3))
        data = results[0]
        self.assertEqual(t1, data)



import numpy as np
from featureflow import Node, Feature

from zounds.nputil import safe_unit_norm
from zounds.timeseries import \
    TimeSlice, Picoseconds, TimeDimension, VariableRateTimeSeries, \
    VariableRateTimeSeriesEncoder, VariableRateTimeSeriesDecoder
from zounds.core import ArrayWithUnits


class MeasureOfTransience(Node):
    """
    Measure of Transience, as defined in section 5.2.1 of
    http://www.mp3-tech.org/programmer/docs/Masri_thesis.pdf

    Uses the ratio of high-frequency content in the signal to detect onsets.
    Effective for percussive onsets.
    """

    def __init__(self, needs=None):
        super(MeasureOfTransience, self).__init__(needs=needs)

    def _first_chunk(self, data):
        data = np.abs(data)
        self._bin_numbers = np.arange(1, data.shape[1] + 1)
        padding = np.zeros(data.shape[1])
        padding[:] = data[0]

        return ArrayWithUnits(
            np.concatenate([padding[None, :], data]), data.dimensions)

    # TODO: this pattern of hanging on to the last sample of the previous chunk,
    # and appending to the next chunk probably happens somewhere else, and
    # should be generalized
    def _enqueue(self, data, pusher):
        if self._cache is None:
            self._cache = data
        else:
            self._cache = self._cache.concatenate(data)

    def _dequeue(self):
        data = self._cache

        self._cache = ArrayWithUnits(
            self._cache[None, -1], self._cache.dimensions)

        return data

    def _process(self, data):
        data = np.abs(data)
        magnitude = (data[:, 2:] ** 2)
        energy = magnitude.sum(axis=1)
        hfc = (magnitude * self._bin_numbers[2:]).sum(axis=1)
        energy[energy == 0] = 1e-12
        hfc[hfc == 0] = 1e-12
        mot = (hfc[1:] / hfc[:-1]) * (hfc[1:] / energy[1:])
        yield mot


class ComplexDomain(Node):
    """
    Complex-domain onset detection as described in
    http://www.eecs.qmul.ac.uk/legacy/dafx03/proceedings/pdfs/dafx81.pdf
    """

    def __init__(self, needs=None):
        super(ComplexDomain, self).__init__(needs=needs)

    def _process(self, data):
        # delta between expected and actual phase
        # TODO: unwrap phases before computing deltas, to avoid artifacts
        # or discontinuties from phase boundary wrapping
        angle = np.angle(data)
        angle = np.unwrap(angle, axis=1)
        angle = np.angle(angle[:, 2] - (2 * angle[:, 1]) + angle[:, 0])

        # expected magnitude
        expected = np.abs(data[:, 1, :])
        # actual magnitude
        actual = np.abs(data[:, 2, :])
        # detection function array
        detect = np.zeros(angle.shape)

        # where phase delta is zero, detection function is the difference 
        # between expected and actual magnitude
        zero_phase_delta_indices = np.where(angle == 0)
        detect[zero_phase_delta_indices] = \
            (expected - actual)[zero_phase_delta_indices]

        # where phase delta is non-zero, detection function combines magnitude
        # and phase deltas
        nonzero_phase_delta_indices = np.where(angle != 0)
        detect[nonzero_phase_delta_indices] = (
            ((expected ** 2) + (actual ** 2) -
             (2 * expected * actual * np.cos(angle))) ** 0.5)[
            nonzero_phase_delta_indices]

        dims = \
            [TimeDimension(data.frequency, data.duration // 3)] \
            + data.dimensions[1:]
        output = ArrayWithUnits(detect.sum(axis=1), dims)
        yield output


class Flux(Node):
    def __init__(self, unit_norm=False, needs=None):
        super(Flux, self).__init__(needs=needs)
        self._memory = None
        self._unit_norm = unit_norm

    def _enqueue(self, data, pusher):
        if self._memory is None:
            self._cache = np.vstack((data[0], data))
        else:
            self._cache = np.vstack((self._memory, data))

        self._cache = ArrayWithUnits(self._cache, data.dimensions)

        self._memory = data[-1]

    def _process(self, data):
        if self._unit_norm:
            data = safe_unit_norm(data)
        diff = np.diff(data, axis=0)

        yield ArrayWithUnits(
            np.linalg.norm(diff, axis=-1), data.dimensions)


class BasePeakPicker(Node):
    def __init__(self, needs=None):
        super(BasePeakPicker, self).__init__(needs=needs)
        self._pos = Picoseconds(0)
        self._leftover_timestamp = self._pos

    def _onset_indices(self, data):
        raise NotImplementedError()

    def _last_chunk(self):
        yield VariableRateTimeSeries((
            (TimeSlice(
             start=self._leftover_timestamp,
             duration=self._pos - self._leftover_timestamp), np.zeros(0)),
        ))

    def _process(self, data):
        td = data.dimensions[0]
        frequency = td.frequency

        indices = self._onset_indices(data)
        timestamps = self._pos + (indices * frequency)
        self._pos += len(data) * frequency

        timestamps = [self._leftover_timestamp] + list(timestamps)
        self._leftover_timestamp = timestamps[-1]

        time_slices = TimeSlice.slices(timestamps)
        vrts = VariableRateTimeSeries([(ts, np.zeros(0)) for ts in time_slices])
        yield vrts


class MovingAveragePeakPicker(BasePeakPicker):
    def __init__(self, aggregate=np.mean, needs=None):
        super(MovingAveragePeakPicker, self).__init__(needs=needs)
        self._aggregate = aggregate

    def _first_chunk(self, data):
        self._center = data.shape[1] // 2
        return data

    def _onset_indices(self, data):
        # compute the threshold for onsets
        agg = self._aggregate(data, axis=1) * 1.25
        # find indices that are peaks
        diff = np.diff(data[:, self._center - 2: self._center + 1])
        peaks = (diff[:, 0] > 0) & (diff[:, 1] < 0)
        # find indices that are above the local average or median
        over_thresh = data[:, self._center] > agg
        # return the intersection of the two
        return np.where(peaks & over_thresh)[0]


class TimeSliceFeature(Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=VariableRateTimeSeriesEncoder,
            decoder=VariableRateTimeSeriesDecoder(),
            **extractor_args):
        super(TimeSliceFeature, self).__init__(
            extractor,
            needs=needs,
            store=store,
            encoder=encoder,
            decoder=decoder,
            key=key,
            **extractor_args)
